"""
TimeSeries Module for QuantumLeap API Client
"""

import logging
import time
from math import inf
from collections import deque
from itertools import count, chain
from typing import Dict, List, Union, Deque, Optional
from urllib.parse import urljoin
import requests
from pydantic import AnyHttpUrl
from pydantic.type_adapter import TypeAdapter
from filip import settings
from filip.clients.base_http_client import BaseHttpClient
from filip.models.base import FiwareHeader
from filip.models.ngsi_v2.subscriptions import Message
from filip.models.ngsi_ld.context import MessageLD
from filip.models.ngsi_v2.timeseries import (
    AggrPeriod,
    AggrMethod,
    AggrScope,
    AttributeValues,
    TimeSeries,
    TimeSeriesHeader,
)
from filip.clients.exceptions import BaseHttpClientException


logger = logging.getLogger(__name__)


class QuantumLeapClient(BaseHttpClient):
    """
    Implements functions to use the FIWARE's QuantumLeap, which subscribes to an
    Orion Context Broker and stores the subscription data in a time series
    database (CrateDB). Further Information:
    https://smartsdk.github.io/ngsi-timeseries-api/#quantumleap
    https://app.swaggerhub.com/apis/heikkilv/quantumleap-api/

    Args:
        url: url of the quantumleap service
        session (Optional):
        fiware_header:
        **kwargs:
    """

    def __init__(
        self,
        url: str = None,
        *,
        session: requests.Session = None,
        fiware_header: FiwareHeader = None,
        **kwargs,
    ):
        # set service url
        url = url or settings.QL_URL
        super().__init__(
            url=url, session=session, fiware_header=fiware_header, **kwargs
        )

    # META API ENDPOINTS
    def get_version(self) -> Dict:
        """
        Gets version of QuantumLeap-Service.

        Returns:
            Dictionary with response
        """
        url = urljoin(self.base_url, "version")
        try:
            res = self.get(url=url, headers=self.headers)
            if res.ok:
                return res.json()
            res.raise_for_status()
        except requests.exceptions.RequestException as err:
            self.logger.error(err)
            raise BaseHttpClientException(
                message=err.response, response=err.response
            ) from err

    def get_health(self) -> Dict:
        """
        This endpoint is intended for administrators of QuantumLeap. Using the
        information returned by this endpoint they can diagnose problems in the
        service or its dependencies. This information is also useful for cloud
        tools such as orchestrators and load balancers with rules based on
        health-checks. Due to the lack of a standardized response format, we
        base the implementation on the draft of
        https://inadarei.github.io/rfc-healthcheck/

        Returns:
            Dictionary with response
        """
        url = urljoin(self.base_url, "health")
        try:
            res = self.get(url=url, headers=self.headers)
            if res.ok:
                return res.json()
            res.raise_for_status()
        except requests.exceptions.RequestException as err:
            self.logger.error(err)
            raise BaseHttpClientException(
                message=err.response.text, response=err.response
            ) from err

    def post_config(self):
        """
        (To Be Implemented) Customize your persistence configuration to
        better suit your needs.
        """
        raise NotImplementedError("Endpoint to be implemented..")

    # INPUT API ENDPOINTS
    def post_notification(self, notification: Union[Message, MessageLD]):
        """
        Notify QuantumLeap the arrival of a new NGSI notification.

        Args:
            notification: Notification Message Object
        """
        url = urljoin(self.base_url, "v2/notify")
        headers = self.headers.copy()
        data = []
        for entity in notification.data:
            data.append(entity.model_dump(exclude_none=True))
        data_set = {"data": data, "subscriptionId": notification.subscriptionId}

        try:
            res = self.post(url=url, headers=headers, json=data_set)
            if res.ok:
                self.logger.debug(res.text)
            else:
                res.raise_for_status()
        except requests.exceptions.RequestException as err:
            msg = (
                f"Could not post notification for subscription id "
                f"{notification.subscriptionId}"
            )
            self.log_error(err=err, msg=msg)
            raise BaseHttpClientException(message=msg, response=err.response) from err

    def post_subscription(
        self,
        cb_url: Union[AnyHttpUrl, str],
        ql_url: Union[AnyHttpUrl, str],
        entity_type: str = None,
        entity_id: str = None,
        id_pattern: str = None,
        attributes: str = None,
        observed_attributes: str = None,
        notified_attributes: str = None,
        throttling: int = None,
        time_index_attribute: str = None,
    ):
        """
        Subscribe QL to process Orion notifications of certain type.
        This endpoint simplifies the creation of the subscription in orion
        that will generate the notifications to be consumed by QuantumLeap in
        order to save historical records. If you want an advanced specification
        of the notifications, you can always create the subscription in orion
        at your will. This endpoint just aims to simplify the common use case.

        Args:
            cb_url:
                url of the context broker
            ql_url:
                The url where Orion can reach QuantumLeap. Do not include
                specific paths.
            entity_type (String):
                The type of entities for which to create a
                subscription, so as to persist historical data of entities of
                this type.
            entity_id (String):
                Id of the entity to track. If specified, it
                takes precedence over the idPattern parameter.
            id_pattern (String): The pattern covering the entity ids for which
                to subscribe. If not specified, QL will track all entities of
                the specified type.
            attributes (String): Comma-separated list of attribute names to
                track.
            observed_attributes (String): Comma-separated list of attribute
                names to track.
            notified_attributes (String): Comma-separated list of attribute
                names to be used to restrict the data of which QL will keep a
                history.
            throttling (int): Minimal period of time in seconds which must
                elapse between two consecutive notifications.
            time_index_attribute (String): The name of a custom attribute to be
                used as a
            time index.
        """
        raise DeprecationWarning(
            "Subscription endpoint of Quantumleap API is "
            "deprecated, use the ORION subscription endpoint "
            "instead"
        )

    def delete_entity(self, entity_id: str, entity_type: Optional[str] = None) -> str:
        """
        Given an entity (with type and id), delete all its historical records.

        Args:
            entity_id (String): Entity id is required.
            entity_type (Optional[String]): Entity type if entity_id alone
                can not uniquely define the entity.

        Raises:
            RequestException, if entity was not found
            Exception, if deleting was not successful

        Returns:
            The entity_id of entity that is deleted.
        """
        url = urljoin(self.base_url, f"v2/entities/{entity_id}")
        headers = self.headers.copy()
        if entity_type is not None:
            params = {"type": entity_type}
        else:
            params = {}

        # The deletion does not always resolves in a success even if an ok is
        # returned.
        # Try to delete multiple times with incrementing waits.
        # If the entity is no longer found the methode returns with a success
        # If the deletion attempt fails after the 10th try, raise an
        # Exception: it could not be deleted
        counter = 0
        while counter < 10:
            self.delete(url=url, headers=headers, params=params)
            try:
                self.get_entity_by_id(entity_id=entity_id, entity_type=entity_type)
            except requests.exceptions.RequestException as err:
                self.logger.info("Entity id '%s' successfully deleted!", entity_id)
                return entity_id
            time.sleep(counter * 5)
            counter += 1

        msg = f"Could not delete QL entity of id {entity_id}"
        logger.error(msg=msg)
        raise Exception(msg)

    def delete_entity_type(self, entity_type: str) -> str:
        """
        Given an entity type, delete all the historical records of all
        entities of such type.
        Args:
            entity_type (String): Type of entities data to be deleted.
        Returns:
            Entity type of the entities deleted.
        """
        url = urljoin(self.base_url, f"v2/types/{entity_type}")
        headers = self.headers.copy()
        try:
            res = self.delete(url=url, headers=headers)
            if res.ok:
                self.logger.info(
                    "Entities of type '%s' successfully deleted!", entity_type
                )
                return entity_type
            res.raise_for_status()
        except requests.exceptions.RequestException as err:
            msg = f"Could not delete entities of type {entity_type}"
            self.log_error(err=err, msg=msg)
            raise BaseHttpClientException(message=msg, response=err.response) from err

    # QUERY API ENDPOINTS
    def __query_builder(
        self,
        url,
        *,
        entity_id: str = None,
        id_pattern: str = None,
        options: str = None,
        entity_type: str = None,
        aggr_method: Union[str, AggrMethod] = None,
        aggr_period: Union[str, AggrPeriod] = None,
        from_date: str = None,
        to_date: str = None,
        last_n: int = None,
        limit: int = 10000,
        offset: int = 0,
        georel: str = None,
        geometry: str = None,
        coords: str = None,
        attrs: str = None,
        aggr_scope: Union[str, AggrScope] = None,
    ) -> Deque[Dict]:
        """
        Private Function to call respective API endpoints, chops large
        requests into multiple single requests and merges the
        responses

        Args:
            url:
            entity_id:
            options:
            entity_type:
            aggr_method:
            aggr_period:
            from_date:
            to_date:
            last_n:
            limit:
                Maximum number of results to retrieve in a single response.
            offset:
                Offset to apply to the response results. For example, if the
                query was to return 10 results and you use an offset of 1, the
                response will return the last 9 values. Make sure you don't
                give more offset than the number of results.
            georel:
            geometry:
            coords:
            attrs:
            aggr_scope:
            id_pattern (str): The pattern covering the entity ids for which
                to subscribe. The pattern follow regular expressions (POSIX
                Extendede) e.g. ".*", "Room.*". Detail information:
                https://en.wikibooks.org/wiki/Regular_Expressions/POSIX-Extended_Regular_Expressions

        Returns:
            Dict
        """
        assert (
            id_pattern is None or entity_id is None
        ), "Cannot have both id and idPattern as parameter."
        params = {}
        headers = self.headers.copy()
        max_records_per_request = 10000
        # create a double ending queue
        res_q: Deque[Dict] = deque([])

        if options:
            params.update({"options": options})
        if entity_type:
            params.update({"type": entity_type})
        if aggr_method:
            aggr_method = AggrMethod(aggr_method)
            params.update({"aggrMethod": aggr_method.value})
        if aggr_period:
            aggr_period = AggrPeriod(aggr_period)
            params.update({"aggrPeriod": aggr_period.value})
        if from_date:
            params.update({"fromDate": from_date})
        if to_date:
            params.update({"toDate": to_date})
        # These values are required for the integrated pagination mechanism
        # maximum items per request
        if limit is None:
            limit = inf
        if offset is None:
            offset = 0
        if georel:
            params.update({"georel": georel})
        if coords:
            params.update({"coords": coords})
        if geometry:
            params.update({"geometry": geometry})
        if attrs:
            params.update({"attrs": attrs})
        if aggr_scope:
            aggr_scope = AggrScope(aggr_scope)
            params.update({"aggr_scope": aggr_scope.value})
        if entity_id:
            params.update({"id": entity_id})
        if id_pattern:
            params.update({"idPattern": id_pattern})

        # This loop will chop large requests into smaller junks.
        # The individual functions will then merge the final response models
        for i in count(0, max_records_per_request):
            try:
                params["offset"] = offset + i

                params["limit"] = min(limit - i, max_records_per_request)
                if params["limit"] <= 0:
                    break

                if last_n:
                    params["lastN"] = min(last_n - i, max_records_per_request)
                    if params["lastN"] <= 0:
                        break

                res = self.get(url=url, params=params, headers=headers)

                if res.ok:
                    self.logger.debug("Received: %s", res.json())

                    # revert append direction when using last_n
                    if last_n:
                        res_q.appendleft(res.json())
                    else:
                        res_q.append(res.json())
                res.raise_for_status()

            except requests.exceptions.RequestException as err:
                if (
                    err.response.status_code == 404
                    and err.response.json().get("error") == "Not Found"
                    and len(res_q) > 0
                ):
                    break
                else:
                    msg = "Could not load entity data"
                    self.log_error(err=err, msg=msg)
                    raise BaseHttpClientException(
                        message=msg, response=err.response
                    ) from err

        self.logger.info("Successfully retrieved entity data")
        return res_q

    # v2/entities
    def get_entities(
        self,
        *,
        entity_type: str = None,
        id_pattern: str = None,
        from_date: str = None,
        to_date: str = None,
        limit: int = 10000,
        offset: int = None,
    ) -> List[TimeSeriesHeader]:
        """
        Get list of all available entities and their context information
        about EntityType and last update date.

        Args:
            entity_type (str): Comma-separated list of entity types whose data
                are to be included in the response. Use only one (no comma)
                when required. If used to resolve ambiguity for the given
                entityId, make sure the given entityId exists for this
                entityType.
            id_pattern (str): The pattern covering the entity ids for which
                to subscribe. The pattern follow regular expressions (POSIX
                Extendede) e.g. ".*", "Room.*". Detail information:
                https://en.wikibooks.org/wiki/Regular_Expressions/POSIX-Extended_Regular_Expressions
            from_date (str): The starting date and time (inclusive) from which
                the context information is queried. Must be in ISO8601 format
                (e.g., 2018-01-05T15:44:34)
            to_date (str): The final date and time (inclusive) from which the
                context information is queried. Must be in ISO8601 format
                (e.g., 2018-01-05T15:44:34).
            limit (int): Maximum number of results to be retrieved.
                Default value : 10000
            offset (int): Offset for the results.

        Returns:
            List of TimeSeriesHeader
        """
        url = urljoin(self.base_url, "v2/entities")
        res = self.__query_builder(
            url=url,
            id_pattern=id_pattern,
            entity_type=entity_type,
            from_date=from_date,
            to_date=to_date,
            limit=limit,
            offset=offset,
        )

        ta = TypeAdapter(List[TimeSeriesHeader])
        return ta.validate_python(res[0])

    # /entities/{entityId}
    def get_entity_by_id(
        self,
        entity_id: str,
        *,
        attrs: str = None,
        entity_type: str = None,
        aggr_method: Union[str, AggrMethod] = None,
        aggr_period: Union[str, AggrPeriod] = None,
        from_date: str = None,
        to_date: str = None,
        last_n: int = None,
        limit: int = 10000,
        offset: int = None,
        georel: str = None,
        geometry: str = None,
        coords: str = None,
        options: str = None,
    ) -> TimeSeries:
        """
        History of N attributes of a given entity instance
        For example, query max water level of the central tank throughout the
        last year. Queries can get more
        sophisticated with the use of filters and query attributes.

        Args:
            entity_id (String): Entity id is required.
            attrs (String):
                Comma-separated list of attribute names whose data are to be
                included in the response. The attributes are retrieved in the
                order specified by this parameter. If not specified, all
                attributes are included in the response in arbitrary order.
            entity_type (String): Comma-separated list of entity types whose
                data are to be included in the response.
            aggr_method (String):
                The function to apply to the raw data filtered by the query
                parameters. If not given, the returned data are the same raw
                inserted data.

                Allowed values: count, sum, avg, min, max
            aggr_period (String):
                If not defined, the aggregation will apply to all the values
                contained in the search result. If defined, the aggregation
                function will instead be applied N times, once for each
                period, and all those results will be considered for the
                response. For example, a query asking for the average
                temperature of an attribute will typically return 1 value.
                However, with an aggregationPeriod of day, you get the daily
                average of the temperature instead (more than one value
                assuming you had measurements across many days within the
                scope of your search result). aggrPeriod must be accompanied
                by an aggrMethod, and the aggrMethod will be applied to all
                the numeric attributes specified in attrs; the rest of the
                non-numerical attrs will be ignored. By default, the response
                is grouped by entity_id. See aggrScope to create aggregation
                across entities:

                Allowed values: year, month, day, hour, minute, second

            from_date (String):
                The starting date and time (inclusive) from which the context
                information is queried. Must be in ISO8601 format (e.g.,
                2018-01-05T15:44:34)
            to_date (String):
                The final date and time (inclusive) from which the context
                information is queried. Must be in ISO8601 format (e.g.,
                2018-01-05T15:44:34)
            last_n (int):
                Used to request only the last N values that satisfy the
                request conditions.
            limit (int): Maximum number of results to be retrieved.
                Default value : 10000
            offset (int):
                Offset to apply to the response results.
            georel (String):
                It specifies a spatial relationship between matching entities
                and a reference shape (geometry). This parameter is used to
                perform geographical queries with the same semantics as in the
                FIWARE-NGSI v2 Specification. Full details can be found in the
                Geographical Queries section of the specification:
                https://fiware.github.io/specifications/ngsiv2/stable/.
            geometry (String):
                Required if georel is specified.  point, line, polygon, box
            coords (String):
                Optional but required if georel is specified. This parameter
                defines the reference shape (geometry) in terms of WGS 84
                coordinates and has the same semantics as in the
                FIWARE-NGSI v2 Specification, except we only accept coordinates
                in decimal degrees---e.g. 40.714,-74.006 is okay, but not
                40 42' 51'',74 0' 21''. Full details can be found in the
                Geographical Queries section of the specification:
                https://fiware.github.io/specifications/ngsiv2/stable/.
            options (String): Key value pair options.

        Returns:
            TimeSeries
        """
        url = urljoin(self.base_url, f"v2/entities/{entity_id}")
        res_q = self.__query_builder(
            url=url,
            attrs=attrs,
            options=options,
            entity_type=entity_type,
            aggr_method=aggr_method,
            aggr_period=aggr_period,
            from_date=from_date,
            to_date=to_date,
            last_n=last_n,
            limit=limit,
            offset=offset,
            georel=georel,
            geometry=geometry,
            coords=coords,
        )
        # merge response chunks
        res = TimeSeries.model_validate(res_q.popleft())
        for item in res_q:
            res.extend(TimeSeries.model_validate(item))

        return res

    # /entities/{entityId}/value
    def get_entity_values_by_id(
        self,
        entity_id: str,
        *,
        attrs: str = None,
        entity_type: str = None,
        aggr_method: Union[str, AggrMethod] = None,
        aggr_period: Union[str, AggrPeriod] = None,
        from_date: str = None,
        to_date: str = None,
        last_n: int = None,
        limit: int = 10000,
        offset: int = None,
        georel: str = None,
        geometry: str = None,
        coords: str = None,
        options: str = None,
    ) -> TimeSeries:
        """
        History of N attributes (values only) of a given entity instance
        For example, query the average pressure, temperature and humidity (
        values only, no metadata) of this
        month in the weather station WS1.

        Args:
            entity_id (String): Entity id is required.
            attrs (String): Comma-separated list of attribute names
            entity_type (String): Comma-separated list of entity types whose
                data are to be included in the response.
            aggr_method (String): The function to apply to the raw data
                filtered. count, sum, avg, min, max
            aggr_period (String): year, month, day, hour, minute, second
            from_date (String): Starting date and time inclusive.
            to_date (String): Final date and time inclusive.
            last_n (int): Request only the last N values.
            limit (int): Maximum number of results to be retrieved.
                Default value : 10000
            offset (int): Offset for the results.
            georel (String): Geographical pattern
            geometry (String): Required if georel is specified.  point, line,
                polygon, box
            coords (String): Required if georel is specified.
                e.g. 40.714,-74.006
            options (String): Key value pair options.

        Returns:
            Response Model
        """
        url = urljoin(self.base_url, f"v2/entities/{entity_id}/value")
        res_q = self.__query_builder(
            url=url,
            attrs=attrs,
            options=options,
            entity_type=entity_type,
            aggr_method=aggr_method,
            aggr_period=aggr_period,
            from_date=from_date,
            to_date=to_date,
            last_n=last_n,
            limit=limit,
            offset=offset,
            georel=georel,
            geometry=geometry,
            coords=coords,
        )

        # merge response chunks
        res = TimeSeries(entityId=entity_id, **res_q.popleft())
        for item in res_q:
            res.extend(TimeSeries(entityId=entity_id, **item))

        return res

    # /entities/{entityId}/attrs/{attrName}
    def get_entity_attr_by_id(
        self,
        entity_id: str,
        attr_name: str,
        *,
        entity_type: str = None,
        aggr_method: Union[str, AggrMethod] = None,
        aggr_period: Union[str, AggrPeriod] = None,
        from_date: str = None,
        to_date: str = None,
        last_n: int = None,
        limit: int = 10000,
        offset: int = None,
        georel: str = None,
        geometry: str = None,
        coords: str = None,
        options: str = None,
    ) -> TimeSeries:
        """
        History of an attribute of a given entity instance
        For example, query max water level of the central tank throughout the
        last year. Queries can get more
        sophisticated with the use of filters and query attributes.

        Args:
            entity_id (String): Entity id is required.
            attr_name (String): The attribute name is required.
            entity_type (String): Comma-separated list of entity types whose
                data are to be included in the response.
            aggr_method (String): The function to apply to the raw data
                filtered. count, sum, avg, min, max
            aggr_period (String): year, month, day, hour, minute, second
            from_date (String): Starting date and time inclusive.
            to_date (String): Final date and time inclusive.
            last_n (int): Request only the last N values.
            limit (int): Maximum number of results to be retrieved.
                Default value : 10000
            offset (int): Offset for the results.
            georel (String): Geographical pattern
            geometry (String): Required if georel is specified.  point, line,
                polygon, box
            coords (String): Required if georel is specified.
                e.g. 40.714,-74.006
            options (String): Key value pair options.

        Returns:
            Response Model
        """
        url = urljoin(self.base_url, f"v2/entities/{entity_id}/attrs" f"/{attr_name}")
        req_q = self.__query_builder(
            url=url,
            entity_id=entity_id,
            options=options,
            entity_type=entity_type,
            aggr_method=aggr_method,
            aggr_period=aggr_period,
            from_date=from_date,
            to_date=to_date,
            last_n=last_n,
            limit=limit,
            offset=offset,
            georel=georel,
            geometry=geometry,
            coords=coords,
        )

        # merge response chunks
        first = req_q.popleft()
        res = TimeSeries(
            entityId=entity_id,
            index=first.get("index"),
            attributes=[AttributeValues(**first)],
        )
        for item in req_q:
            res.extend(
                TimeSeries(
                    entityId=entity_id,
                    index=item.get("index"),
                    attributes=[AttributeValues(**item)],
                )
            )

        return res

    # /entities/{entityId}/attrs/{attrName}/value
    def get_entity_attr_values_by_id(
        self,
        entity_id: str,
        attr_name: str,
        *,
        entity_type: str = None,
        aggr_method: Union[str, AggrMethod] = None,
        aggr_period: Union[str, AggrPeriod] = None,
        from_date: str = None,
        to_date: str = None,
        last_n: int = None,
        limit: int = 10000,
        offset: int = None,
        georel: str = None,
        geometry: str = None,
        coords: str = None,
        options: str = None,
    ) -> TimeSeries:
        """
        History of an attribute (values only) of a given entity instance
        Similar to the previous, but focusing on the values regardless of the
        metadata.

        Args:
            entity_id (String): Entity id is required.
            attr_name (String): The attribute name is required.
            entity_type (String): Comma-separated list of entity types whose
                data are to be included in the response.
            aggr_method (String): The function to apply to the raw data
                filtered. count, sum, avg, min, max
            aggr_period (String): year, month, day, hour, minute, second
            from_date (String): Starting date and time inclusive.
            to_date (String): Final date and time inclusive.
            last_n (int): Request only the last N values.
            limit (int): Maximum number of results to be retrieved.
                Default value : 10000
            offset (int): Offset for the results.
            georel (String): Geographical pattern
            geometry (String): Required if georel is specified.  point, line,
                polygon, box
            coords (String): Required if georel is specified.
                e.g. 40.714,-74.006
            options (String): Key value pair options.

        Returns:
            Response Model
        """
        url = urljoin(
            self.base_url, f"v2/entities/{entity_id}/attrs" f"/{attr_name}/value"
        )
        res_q = self.__query_builder(
            url=url,
            options=options,
            entity_type=entity_type,
            aggr_method=aggr_method,
            aggr_period=aggr_period,
            from_date=from_date,
            to_date=to_date,
            last_n=last_n,
            limit=limit,
            offset=offset,
            georel=georel,
            geometry=geometry,
            coords=coords,
        )
        # merge response chunks
        first = res_q.popleft()
        res = TimeSeries(
            entityId=entity_id,
            index=first.get("index"),
            attributes=[
                AttributeValues(attrName=attr_name, values=first.get("values"))
            ],
        )
        for item in res_q:
            res.extend(
                TimeSeries(
                    entityId=entity_id,
                    index=item.get("index"),
                    attributes=[
                        AttributeValues(attrName=attr_name, values=item.get("values"))
                    ],
                )
            )

        return res

    # /types/{entityType}
    def get_entity_by_type(
        self,
        entity_type: str,
        *,
        attrs: str = None,
        entity_id: str = None,
        id_pattern: str = None,
        aggr_method: Union[str, AggrMethod] = None,
        aggr_period: Union[str, AggrPeriod] = None,
        from_date: str = None,
        to_date: str = None,
        last_n: int = None,
        limit: int = 10000,
        offset: int = None,
        georel: str = None,
        geometry: str = None,
        coords: str = None,
        options: str = None,
        aggr_scope: Union[str, AggrScope] = None,
    ) -> List[TimeSeries]:
        """
        History of N attributes of N entities of the same type.
        For example, query the average pressure, temperature and humidity of
        this month in all the weather stations.
        """
        url = urljoin(self.base_url, f"v2/types/{entity_type}")
        res_q = self.__query_builder(
            url=url,
            entity_id=entity_id,
            id_pattern=id_pattern,
            attrs=attrs,
            options=options,
            aggr_method=aggr_method,
            aggr_period=aggr_period,
            from_date=from_date,
            to_date=to_date,
            last_n=last_n,
            limit=limit,
            offset=offset,
            georel=georel,
            geometry=geometry,
            coords=coords,
            aggr_scope=aggr_scope,
        )

        # merge chunks of response
        res = [
            TimeSeries(entityType=entity_type, **item)
            for item in res_q.popleft().get("entities")
        ]

        for chunk in res_q:
            chunk = [
                TimeSeries(entityType=entity_type, **item)
                for item in chunk.get("entities")
            ]
            for new, old in zip(chunk, res):
                old.extend(new)

        return res

    # /types/{entityType}/value
    def get_entity_values_by_type(
        self,
        entity_type: str,
        *,
        attrs: str = None,
        entity_id: str = None,
        id_pattern: str = None,
        aggr_method: Union[str, AggrMethod] = None,
        aggr_period: Union[str, AggrPeriod] = None,
        from_date: str = None,
        to_date: str = None,
        last_n: int = None,
        limit: int = 10000,
        offset: int = None,
        georel: str = None,
        geometry: str = None,
        coords: str = None,
        options: str = None,
        aggr_scope: Union[str, AggrScope] = None,
    ) -> List[TimeSeries]:
        """
        History of N attributes (values only) of N entities of the same type.
        For example, query the average pressure, temperature and humidity (
        values only, no metadata) of this month in
        all the weather stations.
        """
        url = urljoin(self.base_url, f"v2/types/{entity_type}/value")
        res_q = self.__query_builder(
            url=url,
            entity_id=entity_id,
            id_pattern=id_pattern,
            attrs=attrs,
            options=options,
            entity_type=entity_type,
            aggr_method=aggr_method,
            aggr_period=aggr_period,
            from_date=from_date,
            to_date=to_date,
            last_n=last_n,
            limit=limit,
            offset=offset,
            georel=georel,
            geometry=geometry,
            coords=coords,
            aggr_scope=aggr_scope,
        )
        # merge chunks of response
        res = [
            TimeSeries(entityType=entity_type, **item)
            for item in res_q.popleft().get("values")
        ]

        for chunk in res_q:
            chunk = [
                TimeSeries(entityType=entity_type, **item)
                for item in chunk.get("values")
            ]
            for new, old in zip(chunk, res):
                old.extend(new)

        return res

    # /types/{entityType}/attrs/{attrName}
    def get_entity_attr_by_type(
        self,
        entity_type: str,
        attr_name: str,
        *,
        entity_id: str = None,
        id_pattern: str = None,
        aggr_method: Union[str, AggrMethod] = None,
        aggr_period: Union[str, AggrPeriod] = None,
        from_date: str = None,
        to_date: str = None,
        last_n: int = None,
        limit: int = 10000,
        offset: int = None,
        georel: str = None,
        geometry: str = None,
        coords: str = None,
        options: str = None,
        aggr_scope: Union[str, AggrScope] = None,
    ) -> List[TimeSeries]:
        """
        History of an attribute of N entities of the same type.
        For example, query the pressure measurements of this month in all the
        weather stations. Note in the response,
        the index and values arrays are parallel. Also, when using
        aggrMethod, the aggregation is done by-entity
        instance. In this case, the index array is just the fromDate and
        toDate values user specified in the request
        (if any).

        Args:
            entity_type (String): Entity type is required.
            attr_name (String): The attribute name is required.
            entity_id (String): Comma-separated list of entity ids whose data
                are to be included in the response.
            aggr_method (String): The function to apply to the raw data
                filtered. count, sum, avg, min, max
            aggr_period (String): year, month, day, hour, minute, second
            aggr_scope (str): Optional. (This parameter is not yet supported).
                When the query results cover historical data for multiple
                entities instances, you can define the aggregation method to be
                applied for each entity instance [entity] or across
                them [global]
            from_date (String): Starting date and time inclusive.
            to_date (String): Final date and time inclusive.
            last_n (int): Request only the last N values.
            limit (int): Maximum number of results to be retrieved.
                Default value : 10000
            offset (int): Offset for the results.
            georel (String): Geographical pattern
            geometry (String): Required if georel is specified.  point, line,
                polygon, box
            coords (String): Required if georel is specified.
                e.g. 40.714,-74.006
            options (String): Key value pair options.

        Returns:
            Response Model
        """
        url = urljoin(self.base_url, f"v2/types/{entity_type}/attrs" f"/{attr_name}")
        res_q = self.__query_builder(
            url=url,
            entity_id=entity_id,
            id_pattern=id_pattern,
            options=options,
            entity_type=entity_type,
            aggr_method=aggr_method,
            aggr_period=aggr_period,
            from_date=from_date,
            to_date=to_date,
            last_n=last_n,
            limit=limit,
            offset=offset,
            georel=georel,
            geometry=geometry,
            coords=coords,
            aggr_scope=aggr_scope,
        )

        # merge chunks of response
        first = res_q.popleft()
        res = [
            TimeSeries(
                index=item.get("index"),
                entityType=entity_type,
                entityId=item.get("entityId"),
                attributes=[
                    AttributeValues(
                        attrName=first.get("attrName"), values=item.get("values")
                    )
                ],
            )
            for item in first.get("entities")
        ]

        for chunk in res_q:
            chunk = [
                TimeSeries(
                    index=item.get("index"),
                    entityType=entity_type,
                    entityId=item.get("entityId"),
                    attributes=[
                        AttributeValues(
                            attrName=chunk.get("attrName"), values=item.get("values")
                        )
                    ],
                )
                for item in chunk.get("entities")
            ]
            for new, old in zip(chunk, res):
                old.extend(new)

        return res

    # /types/{entityType}/attrs/{attrName}/value
    def get_entity_attr_values_by_type(
        self,
        entity_type: str,
        attr_name: str,
        *,
        entity_id: str = None,
        id_pattern: str = None,
        aggr_method: Union[str, AggrMethod] = None,
        aggr_period: Union[str, AggrPeriod] = None,
        from_date: str = None,
        to_date: str = None,
        last_n: int = None,
        limit: int = 10000,
        offset: int = None,
        georel: str = None,
        geometry: str = None,
        coords: str = None,
        options: str = None,
        aggr_scope: Union[str, AggrScope] = None,
    ) -> List[TimeSeries]:
        """
        History of an attribute (values only) of N entities of the same type.
        For example, query the average pressure (values only, no metadata) of
        this month in all the weather stations.

        Args:
            aggr_scope:
            entity_type (String): Entity type is required.
            attr_name (String): The attribute name is required.
            entity_id (String): Comma-separated list of entity ids whose data
                are to be included in the response.
            aggr_method (String): The function to apply to the raw data
                filtered. count, sum, avg, min, max
            aggr_period (String): year, month, day, hour, minute, second
            aggr_scope (String):
            from_date (String): Starting date and time inclusive.
            to_date (String): Final date and time inclusive.
            last_n (int): Request only the last N values.
            limit (int): Maximum number of results to be retrieved.
                Default value : 10000
            offset (int): Offset for the results.
            georel (String): Geographical pattern
            geometry (String): Required if georel is specified.  point, line,
                polygon, box
            coords (String): Required if georel is specified.
                e.g. 40.714,-74.006
            options (String): Key value pair options.

        Returns:
            Response Model
        """
        url = urljoin(
            self.base_url, f"v2/types/{entity_type}/attrs/" f"{attr_name}/value"
        )
        res_q = self.__query_builder(
            url=url,
            entity_id=entity_id,
            id_pattern=id_pattern,
            options=options,
            entity_type=entity_type,
            aggr_method=aggr_method,
            aggr_period=aggr_period,
            from_date=from_date,
            to_date=to_date,
            last_n=last_n,
            limit=limit,
            offset=offset,
            georel=georel,
            geometry=geometry,
            coords=coords,
            aggr_scope=aggr_scope,
        )

        # merge chunks of response
        res = [
            TimeSeries(
                index=item.get("index"),
                entityType=entity_type,
                entityId=item.get("entityId"),
                attributes=[
                    AttributeValues(attrName=attr_name, values=item.get("values"))
                ],
            )
            for item in res_q.popleft().get("values")
        ]

        for chunk in res_q:
            chunk = [
                TimeSeries(
                    index=item.get("index"),
                    entityType=entity_type,
                    entityId=item.get("entityId"),
                    attributes=[
                        AttributeValues(attrName=attr_name, values=item.get("values"))
                    ],
                )
                for item in chunk.get("values")
            ]

            for new, old in zip(chunk, res):
                old.extend(new)
        return res

    # v2/attrs
    def get_entity_by_attrs(
        self,
        *,
        entity_type: str = None,
        from_date: str = None,
        to_date: str = None,
        limit: int = 10000,
        offset: int = None,
    ) -> List[TimeSeries]:
        """
        Get list of timeseries data grouped by each existing attribute name.
        The timeseries data include all entities corresponding to each
        attribute name as well as the index and values of this attribute in
        this entity.

        Args:
            entity_type (str): Comma-separated list of entity types whose data
                are to be included in the response. Use only one (no comma)
                when required. If used to resolve ambiguity for the given
                entityId, make sure the given entityId exists for this
                entityType.
            from_date (str): The starting date and time (inclusive) from which
                the context information is queried. Must be in ISO8601 format
                (e.g., 2018-01-05T15:44:34)
            to_date (str): The final date and time (inclusive) from which the
                context information is queried. Must be in ISO8601 format
                (e.g., 2018-01-05T15:44:34).
            limit (int): Maximum number of results to be retrieved.
                Default value : 10000
            offset (int): Offset for the results.

        Returns:
            List of TimeSeriesEntities
        """
        url = urljoin(self.base_url, "v2/attrs")
        res_q = self.__query_builder(
            url=url,
            entity_type=entity_type,
            from_date=from_date,
            to_date=to_date,
            limit=limit,
            offset=offset,
        )
        first = res_q.popleft()

        res = chain.from_iterable(
            map(lambda x: self.transform_attr_response_model(x), first.get("attrs"))
        )
        for chunk in res_q:
            chunk = chain.from_iterable(
                map(lambda x: self.transform_attr_response_model(x), chunk.get("attrs"))
            )

            for new, old in zip(chunk, res):
                old.extend(new)

        return list(res)

    # v2/attrs/{attr_name}
    def get_entity_by_attr_name(
        self,
        *,
        attr_name: str,
        entity_type: str = None,
        from_date: str = None,
        to_date: str = None,
        limit: int = 10000,
        offset: int = None,
    ) -> List[TimeSeries]:
        """
        Get list of all entities containing this attribute name, as well as
        getting the index and values of this attribute in every corresponding
        entity.

        Args:
            attr_name (str): The attribute name in interest.
            entity_type (str): Comma-separated list of entity types whose data
                are to be included in the response. Use only one (no comma)
                when required. If used to resolve ambiguity for the given
                entityId, make sure the given entityId exists for this
                entityType.
            from_date (str): The starting date and time (inclusive) from which
                the context information is queried. Must be in ISO8601 format
                (e.g., 2018-01-05T15:44:34)
            to_date (str): The final date and time (inclusive) from which the
                context information is queried. Must be in ISO8601 format
                (e.g., 2018-01-05T15:44:34).
            limit (int): Maximum number of results to be retrieved.
                Default value : 10000
            offset (int): Offset for the results.

        Returns:
            List of TimeSeries
        """
        url = urljoin(self.base_url, f"/v2/attrs/{attr_name}")
        res_q = self.__query_builder(
            url=url,
            entity_type=entity_type,
            from_date=from_date,
            to_date=to_date,
            limit=limit,
            offset=offset,
        )

        first = res_q.popleft()
        res = self.transform_attr_response_model(first)

        for chunk in res_q:
            chunk = self.transform_attr_response_model(chunk)
            for new, old in zip(chunk, res):
                old.extend(new)
        return list(res)

    def transform_attr_response_model(self, attr_response):
        res = []
        attr_name = attr_response.get("attrName")
        for entity_group in attr_response.get("types"):
            timeseries = map(
                lambda entity: TimeSeries(
                    entityId=entity.get("entityId"),
                    entityType=entity_group.get("entityType"),
                    index=entity.get("index"),
                    attributes=[
                        AttributeValues(attrName=attr_name, values=entity.get("values"))
                    ],
                ),
                entity_group.get("entities"),
            )
            res.append(timeseries)
        return chain.from_iterable(res)



"""
Abstract class for all IoTA MQTT message encoders
"""

import logging
from abc import ABC
from datetime import datetime
from typing import Dict, Tuple
from paho.mqtt.client import MQTTMessage
from filip.models.mqtt import IoTAMQTTMessageType
from filip.utils import convert_datetime_to_iso_8601_with_z_suffix


class BaseEncoder(ABC):
    """
    Abstract class for all IoTA MQTT message encoders
    """

    prefix: str = ""

    def __init__(self):
        # setup logging functionality
        self.logger = logging.getLogger(
            name=f"{self.__class__.__module__}." f"{self.__class__.__name__}"
        )
        self.logger.addHandler(logging.NullHandler())

    def decode_message(
        self, msg: MQTTMessage, decoder: str = "utf-8"
    ) -> Tuple[str, str, str]:
        """
        Decode message for ingoing traffic
        Args:
            msg: Message class
            decoder: encoding identifier

        Returns:
            apikey
            device_id
            payload
        """
        topic = msg.topic.strip("/")
        topic = topic.split("/")
        apikey = None
        device_id = None
        payload = msg.payload.decode(decoder)
        if topic[-1] == "cmd":
            apikey = topic[0]
            device_id = topic[1]

        if any((apikey, device_id, payload)) is None:
            raise ValueError

        return apikey, device_id, payload

    def encode_msg(
        self, device_id: str, payload: Dict, msg_type: IoTAMQTTMessageType
    ) -> str:
        """
        Encode message for outgoing traffic

        Args:
            device_id: id of the iot device
            payload: payload to send
            msg_type: kind of message to send

        """
        raise NotImplementedError

    @classmethod
    def _parse_timestamp(cls, payload: Dict) -> Dict:
        """
        Helper function to parse timestamps

        Args:
            payload: payload to reformat

        Returns:
            Dictionary containing the formatted payload
        """
        if payload.get("timeInstant", None):
            timestamp = payload["timeInstant"]
            if isinstance(timestamp, str):
                timestamp = datetime.fromisoformat(payload["timeInstant"])
            if isinstance(timestamp, datetime):
                payload["timeInstant"] = convert_datetime_to_iso_8601_with_z_suffix(
                    payload["timeInstant"]
                )
            else:
                raise ValueError("Not able to parse datetime")
        return payload

    @classmethod
    def _raise_encoding_error(cls, payload: Dict, msg_type: IoTAMQTTMessageType):
        """
        Helper function to provide consistent error messages
        Args:
            payload: Invalid message content
            msg_type: Invalid message type

        Returns:
            None

        Raises:
            ValueError
        """
        ValueError(
            f"Message format not supported! \n "
            f"Message Type: {msg_type} \n "
            f"Payload: {payload}"
        )



"""
Test module for logging functionality
"""

import logging
from unittest import TestCase


class TestLoggingConfig(TestCase):
    """
    Test case for logging functionality
    """

    def setUp(self) -> None:
        self.logger = logging.getLogger(name=f"{__package__}.{self.__class__.__name__}")

    def test_overwrite_config(self):
        """
        Tests to overwrite the logger configurations

        Returns:
            None
        """

        # Try to set logging level before calling logging.basisConfig()
        # Since no handler is configured this will fail
        self.logger = logging.getLogger(name=f"{__package__}.{self.__class__.__name__}")
        self.logger.warning(
            "Trying to set log_level to '%s' via settings "
            "before calling basicConfig. This will fail "
            "because no handler is added to the logger",
            logging.DEBUG,
        )
        self.logger.setLevel(level=logging.DEBUG)
        # The next line will not show up!
        self.logger.info("Current LOG_LEVEL is '%s'", self.logger.level)

        # Try to set logging level before calling logging.basisConfig()
        # but adding a handler before.
        self.logger.handlers.clear()
        handler = logging.StreamHandler()
        formatter = logging.Formatter(
            fmt="Custom Logging Stream %(asctime)s - "
            "%(name)s - "
            "%(levelname)s : %(message)s"
        )
        handler.setFormatter(formatter)
        self.logger.info(
            "Set LOG_LEVEL to '%s' via settings and adding a " "handler before. ",
            logging.DEBUG,
        )
        self.logger.addHandler(handler)
        self.logger.setLevel(level=logging.DEBUG)
        self.logger.info("Current LOG_LEVEL has changed to '%s'", self.logger.level)
        # The next line will not show up!
        self.logger.debug("Current LOG_LEVEL is '%s'", self.logger.level)

        # Try to set loglevel via basicConfig
        new_loglevel = logging.WARNING
        self.logger.info("Set LOG_LEVEL to '%s' via basicConfig", new_loglevel)
        # logging.basicConfig(level=new_loglevel,
        #                    format='%(asctime)s : %(name)s : %(levelname)s : '
        #                           '%(message)s')
        logger = logging.getLogger()
        handler = logging.StreamHandler()
        formatter = logging.Formatter(
            fmt="Root Stream %(asctime)s - " "%(name)s - " "%(levelname)s : %(message)s"
        )
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel("DEBUG")
        logger.debug("Initialize root logger")
        # This message will show up twice because now a handler is configured
        # in the root logger and all messages will be forwarded
        # but it does because the
        # loggers are detached and we need to delete the handler first
        self.logger.info(
            "Current LOG_LEVEL '%s' (this will appear twice)", self.logger.level
        )

    def tearDown(self) -> None:
        pass



"""
Settings module to set url from .env.filip file. This can also seen as an
example for other applications such as webapp that use the library. Using
`*.env` belongs to best practices in containerized applications. Pydantic
provides a convenient and clean way to manage environments.
"""

from pydantic import Field, AnyHttpUrl, AliasChoices, AnyUrl
from pydantic_settings import BaseSettings, SettingsConfigDict
from pathlib import Path
import os
from dotenv import find_dotenv

ROOT_DIR = os.path.realpath(os.path.join(os.path.dirname(__file__), ".."))


class Settings(BaseSettings):
    """
    Settings class that reads environment variables from a local `.env.filip`
    file or environment variables. The `.env.filip` can be located anywhere
    in the FiLiP repository.
    """

    model_config = SettingsConfigDict(
        env_file=find_dotenv(Path(ROOT_DIR) / ".env.filip"),
        env_file_encoding="utf-8",
        case_sensitive=False,
        extra="ignore",
    )

    CB_URL: AnyHttpUrl = Field(
        default="http://127.0.0.1:1026",
        validation_alias=AliasChoices(
            "ORION_URL", "CB_URL", "CB_HOST", "CONTEXTBROKER_URL", "OCB_URL"
        ),
    )
    LD_CB_URL: AnyHttpUrl = Field(
        default="http://127.0.0.1:1027",
        validation_alias=AliasChoices(
            "LD_ORION_URL", "LD_CB_URL", "ORION_LD_URL", "SCORPIO_URL", "STELLIO_URL"
        ),
    )

    IOTA_URL: AnyHttpUrl = Field(
        default="http://127.0.0.1:4041", validation_alias="IOTA_URL"
    )

    QL_URL: AnyHttpUrl = Field(
        default="http://127.0.0.1:8668",
        validation_alias=AliasChoices("QUANTUMLEAP_URL", "QL_URL"),
    )

    MQTT_BROKER_URL: AnyUrl = Field(
        default="mqtt://127.0.0.1:1883",
        validation_alias=AliasChoices("MQTT_BROKER_URL", "MQTT_URL", "MQTT_BROKER"),
    )
    LD_MQTT_BROKER_URL: AnyUrl = Field(
        default="mqtt://127.0.0.1:1884",
        validation_alias=AliasChoices(
            "LD_MQTT_BROKER_URL", "LD_MQTT_URL", "LD_MQTT_BROKER"
        ),
    )

    FIWARE_SERVICE: str = Field(
        default="filip", validation_alias=AliasChoices("FIWARE_SERVICE")
    )


# create settings object
settings = Settings()



"""
Module contains models for accessing and interaction with FIWARE's IoT-Agents.
"""

from __future__ import annotations
import logging
import itertools
import warnings
from enum import Enum
from typing import Any, Dict, Optional, List, Union
import pytz
from pydantic import (
    field_validator,
    model_validator,
    ConfigDict,
    BaseModel,
    Field,
    AnyHttpUrl,
)
from filip.models.base import NgsiVersion, DataType
from filip.models.ngsi_v2.base import (
    BaseAttribute,
    BaseValueAttribute,
    BaseNameAttribute,
)
from filip.utils.validators import (
    validate_fiware_datatype_string_protect,
    validate_fiware_datatype_standard,
    validate_jexl_expression,
    validate_expression_language,
    validate_fiware_attribute_name_regex,
)

logger = logging.getLogger()


class ExpressionLanguage(str, Enum):
    """
    Options for expression language
    """

    LEGACY = "legacy"
    JEXL = "jexl"


class PayloadProtocol(str, Enum):
    """
    Options for payload protocols
    """

    IOTA_JSON = "IoTA-JSON"
    IOTA_UL = "PDI-IoTA-UltraLight"
    LORAWAN = "LoRaWAN"


class TransportProtocol(str, Enum):
    """
    Options for transport protocols
    """

    MQTT = "MQTT"
    AMQP = "AMQP"
    HTTP = "HTTP"


class IoTABaseAttribute(BaseAttribute, BaseNameAttribute):
    """
    Base model for device attributes
    """

    expression: Optional[str] = Field(
        default=None,
        description="indicates that the value of the target attribute will "
        "not be the plain value or the measurement, but an "
        "expression based on a combination of the reported values. "
        "See the Expression Language definition for details "
        "(https://iotagent-node-lib.readthedocs.io/en/latest/"
        "api.html#expression-language-support)",
    )
    entity_name: Optional[str] = Field(
        default=None,
        description="entity_name: the presence of this attribute indicates "
        "that the value will not be stored in the original device "
        "entity but in a new entity with an ID given by this "
        "attribute. The type of this additional entity can be "
        "configured with the entity_type attribute. If no type is "
        "configured, the device entity type is used instead. "
        "Entity names can be defined as expressions, using the "
        "Expression Language definition "
        "(https://iotagent-node-lib.readthedocs.io/en/latest/"
        "api.html#expression-language-support). Allowed characters are"
        " the ones in the plain ASCII set, except the following "
        "ones: control characters, whitespace, &, ?, / and #.",
        max_length=256,
        min_length=1,
    )
    valid_entity_name = field_validator("entity_name")(
        validate_fiware_datatype_standard
    )
    entity_type: Optional[str] = Field(
        default=None,
        description="configures the type of an alternative entity. "
        "Allowed characters "
        "are the ones in the plain ASCII set, except the following "
        "ones: control characters, whitespace, &, ?, / and #.",
        max_length=256,
        min_length=1,
    )
    valid_entity_type = field_validator("entity_type")(
        validate_fiware_datatype_standard
    )
    reverse: Optional[str] = Field(
        default=None,
        description="add bidirectionality expressions to the attribute. See "
        "the bidirectionality transformation plugin in the "
        "Data Mapping Plugins section for details. "
        "(https://iotagent-node-lib.readthedocs.io/en/latest/api/"
        "index.html#data-mapping-plugins)",
    )

    def __eq__(self, other):
        if isinstance(other, BaseAttribute):
            return self.name == other.name
        else:
            return self.model_dump() == other


class DeviceAttribute(IoTABaseAttribute):
    """
    Model for active device attributes
    """

    object_id: Optional[str] = Field(
        default=None, description="name of the attribute as coming from the device."
    )


class LazyDeviceAttribute(BaseNameAttribute):
    """
    Model for lazy device attributes
    """

    type: Union[DataType, str] = Field(
        default=DataType.TEXT,
        description="The attribute type represents the NGSI value type of the "
        "attribute value. Note that FIWARE NGSI has its own type "
        "system for attribute values, so NGSI value types are not "
        "the same as JSON types. Allowed characters "
        "are the ones in the plain ASCII set, except the following "
        "ones: control characters, whitespace, &, ?, / and #.",
        max_length=256,
        min_length=1,
    )
    valid_type = field_validator("type")(validate_fiware_datatype_string_protect)


class DeviceCommand(BaseModel):
    """
    Model for commands
    """

    name: str = Field(
        description="ID of the attribute in the target entity in the "
        "Context Broker. Allowed characters "
        "are the ones in the plain ASCII set, except the following "
        "ones: control characters, whitespace, &, ?, / and #.",
        max_length=256,
        min_length=1,
    )
    valid_name = field_validator("name")(validate_fiware_attribute_name_regex)
    type: Union[DataType, str] = Field(
        description="name of the type of the attribute in the target entity. ",
        default=DataType.COMMAND,
    )


class StaticDeviceAttribute(IoTABaseAttribute, BaseValueAttribute):
    """
    Model for static device attributes
    """

    pass


class ServiceGroup(BaseModel):
    """
    Model for device service group.
    https://iotagent-node-lib.readthedocs.io/en/latest/api/index.html#service-group-api
    """

    service: Optional[str] = Field(
        default=None, description="ServiceGroup of the devices of this type"
    )
    subservice: Optional[str] = Field(
        default=None,
        description="Subservice of the devices of this type.",
        pattern="^/",
    )
    resource: str = Field(
        description="string representing the Southbound resource that will be "
        "used to assign a type to a device  (e.g.: pathname in the "
        "southbound port)."
    )
    apikey: str = Field(
        description="API Key string. It is a key used for devices belonging "
        "to this service_group. If "
        ", service_group does not use "
        "apikey, but it must be specified."
    )
    timestamp: Optional[bool] = Field(
        default=None,
        description="Optional flag about whether or not to add the TimeInstant "
        "attribute to the device entity created, as well as a "
        "TimeInstant metadata to each attribute, with the current "
        "timestamp. With NGSI-LD, the Standard observedAt "
        "property-of-a-property is created instead.",
    )
    entity_type: Optional[str] = Field(
        default=None,
        description="name of the Entity type to assign to the group. "
        "Allowed characters "
        "are the ones in the plain ASCII set, except the following "
        "ones: control characters, whitespace, &, ?, / and #.",
        max_length=256,
        min_length=1,
    )
    valid_entity_type = field_validator("entity_type")(
        validate_fiware_datatype_standard
    )
    trust: Optional[str] = Field(
        default=None,
        description="trust token to use for secured access to the "
        "Context Broker for this type of devices (optional; only "
        "needed for secured scenarios).",
    )
    cbHost: Optional[AnyHttpUrl] = Field(
        default=None,
        description="Context Broker connection information. This options can "
        "be used to override the global ones for specific types of "
        "devices.",
    )

    @field_validator("cbHost")
    @classmethod
    def validate_cbHost(cls, value):
        """
        convert cbHost to str
        Returns:
            timezone
        """
        return str(value) if value else value

    lazy: Optional[List[LazyDeviceAttribute]] = Field(
        default=[],
        description="list of common lazy attributes of the device. For each "
        "attribute, its name and type must be provided.",
    )
    commands: Optional[List[DeviceCommand]] = Field(
        default=[],
        description="list of common commands attributes of the device. For each "
        "attribute, its name and type must be provided, additional "
        "metadata is optional",
    )
    attributes: Optional[List[DeviceAttribute]] = Field(
        default=[],
        description="list of common commands attributes of the device. For "
        "each attribute, its name and type must be provided, "
        "additional metadata is optional.",
    )
    static_attributes: Optional[List[StaticDeviceAttribute]] = Field(
        default=[],
        description="this attributes will be added to all the entities of this "
        "group 'as is', additional metadata is optional.",
    )
    internal_attributes: Optional[List[Dict[str, Any]]] = Field(
        default=[],
        description="optional section with free format, to allow specific "
        "IoT Agents to store information along with the devices "
        "in the Device Registry.",
    )
    expressionLanguage: Optional[ExpressionLanguage] = Field(
        default=ExpressionLanguage.JEXL,
        description="optional boolean value, to set expression language used "
        "to compute expressions, possible values are: "
        "legacy or jexl, but legacy is deprecated. If it is set None, jexl is used.",
    )
    valid_expressionLanguage = field_validator("expressionLanguage")(
        validate_expression_language
    )
    explicitAttrs: Optional[bool] = Field(
        default=False,
        description="optional boolean value, to support selective ignore "
        "of measures so that IOTA does not progress. If not "
        "specified default is false.",
    )
    autoprovision: Optional[bool] = Field(
        default=True,
        description="optional boolean: If false, autoprovisioned devices "
        "(i.e. devices that are not created with an explicit "
        "provision operation but when the first measure arrives) "
        "are not allowed in this group. "
        "Default (in the case of omitting the field) is true.",
    )
    ngsiVersion: Optional[NgsiVersion] = Field(
        default="v2",
        description="optional string value used in mixed mode to switch between"
        " NGSI-v2 and NGSI-LD payloads. Possible values are: "
        "v2 or ld. The default is v2. When not running in mixed "
        "mode, this field is ignored.",
    )
    defaultEntityNameConjunction: Optional[str] = Field(
        default=None,
        description="optional string value to set default conjunction string "
        "used to compose a default entity_name when is not "
        "provided at device provisioning time.",
    )


class DeviceSettings(BaseModel):
    """
    Model for iot device settings
    """

    model_config = ConfigDict(validate_assignment=True)
    timezone: Optional[str] = Field(
        default="Europe/London", description="Time zone of the sensor if it has any"
    )
    timestamp: Optional[bool] = Field(
        default=None,
        description="Optional flag about whether or not to add the TimeInstant "
        "attribute to the device entity created, as well as a "
        "TimeInstant metadata to each attribute, with the current "
        "timestamp. With NGSI-LD, the Standard observedAt "
        "property-of-a-property is created instead.",
    )
    apikey: Optional[str] = Field(
        default=None,
        description="Optional Apikey key string to use instead of group apikey",
    )
    endpoint: Optional[AnyHttpUrl] = Field(
        default=None,
        description="Endpoint where the device is going to receive commands, "
        "if any.",
    )
    protocol: Optional[Union[PayloadProtocol, str]] = Field(
        default=None,
        description="Name of the device protocol, for its use with an " "IoT Manager.",
    )
    transport: Optional[Union[TransportProtocol, str]] = Field(
        default=None,
        description="Name of the device transport protocol, for the IoT Agents "
        "with multiple transport protocols.",
    )
    expressionLanguage: Optional[ExpressionLanguage] = Field(
        default=ExpressionLanguage.JEXL,
        description="optional boolean value, to set expression language used "
        "to compute expressions, possible values are: "
        "legacy or jexl, but legacy is deprecated. If it is set None, jexl is used.",
    )
    valid_expressionLanguage = field_validator("expressionLanguage")(
        validate_expression_language
    )
    explicitAttrs: Optional[bool] = Field(
        default=False,
        description="optional boolean value, to support selective ignore "
        "of measures so that IOTA does not progress. If not "
        "specified default is false.",
    )


class Device(DeviceSettings):
    """
    Model for iot devices.
    https://iotagent-node-lib.readthedocs.io/en/latest/api/index.html#device-api
    """

    model_config = ConfigDict(validate_default=True, validate_assignment=True)
    device_id: str = Field(
        description="Device ID that will be used to identify the device"
    )
    service: Optional[str] = Field(
        default=None,
        description="Name of the service the device belongs to "
        "(will be used in the fiware-service header).",
        max_length=50,
    )
    service_path: Optional[str] = Field(
        default="/",
        description="Name of the subservice the device belongs to "
        "(used in the fiware-servicepath header).",
        max_length=51,
        pattern="^/",
    )
    entity_name: str = Field(
        description="Name of the entity representing the device in "
        "the Context Broker Allowed characters "
        "are the ones in the plain ASCII set, except the following "
        "ones: control characters, whitespace, &, ?, / and #.",
        max_length=256,
        min_length=1,
    )
    valid_entity_name = field_validator("entity_name")(
        validate_fiware_datatype_standard
    )
    entity_type: str = Field(
        description="Type of the entity in the Context Broker. "
        "Allowed characters "
        "are the ones in the plain ASCII set, except the following "
        "ones: control characters, whitespace, &, ?, / and #.",
        max_length=256,
        min_length=1,
    )
    valid_entity_type = field_validator("entity_type")(
        validate_fiware_datatype_standard
    )
    lazy: List[LazyDeviceAttribute] = Field(
        default=[], description="List of lazy attributes of the device"
    )
    commands: List[DeviceCommand] = Field(
        default=[], description="List of commands of the device"
    )
    attributes: List[DeviceAttribute] = Field(
        default=[], description="List of active attributes of the device"
    )
    static_attributes: Optional[List[StaticDeviceAttribute]] = Field(
        default=[],
        description="List of static attributes to append to the entity. All the"
        " updateContext requests to the CB will have this set of "
        "attributes appended.",
    )
    internal_attributes: Optional[List[Dict[str, Any]]] = Field(
        default=[],
        description="List of internal attributes with free format for specific "
        "IoT Agent configuration",
    )
    ngsiVersion: NgsiVersion = Field(
        default=NgsiVersion.v2,
        description="optional string value used in mixed mode to switch between"
        " NGSI-v2 and NGSI-LD payloads. Possible values are: "
        "v2 or ld. The default is v2. When not running in "
        "mixed mode, this field is ignored.",
    )

    @field_validator("timezone")
    @classmethod
    def validate_timezone(cls, value):
        """
        validate timezone
        Returns:
            timezone
        """
        assert value in pytz.all_timezones
        return value

    @model_validator(mode="after")
    def validate_device_attributes_expression(self):
        """
        Validates device attributes expressions based on the expression language (JEXL or Legacy, where Legacy is
        deprecated).

        Args:
            self: The Device instance.

        Returns:
            The Device instance after validation.
        """
        if self.expressionLanguage == ExpressionLanguage.JEXL:
            for attribute in self.attributes:
                if attribute.expression:
                    validate_jexl_expression(
                        attribute.expression, attribute.name, self.device_id
                    )
        elif self.expressionLanguage == ExpressionLanguage.LEGACY:
            warnings.warn(
                f"No validation for legacy expression language of Device {self.device_id}."
            )

        return self

    @model_validator(mode="after")
    def validate_duplicated_device_attributes(self):
        """
        Check whether device has identical attributes
        Args:
            self: dict of Device instance.

        Returns:
            The dict of Device instance after validation.
        """
        for i, attr in enumerate(self.attributes):
            for other_attr in self.attributes[:i] + self.attributes[i + 1 :]:
                if attr.model_dump() == other_attr.model_dump():
                    raise ValueError(f"Duplicated attributes found: {attr.name}")
        return self

    @model_validator(mode="after")
    def validate_device_attributes_name_object_id(self):
        """
        Validate the device regarding the behavior with devices attributes.
        According to https://iotagent-node-lib.readthedocs.io/en/latest/api.html and
        based on our best practice, following rules are checked
            - name is required, but not necessarily unique
            - object_id is not required, if given must be unique, i.e. not equal to any
                existing object_id and name
        Args:
            self: dict of Device instance.

        Returns:
            The dict of Device instance after validation.
        """
        for i, attr in enumerate(self.attributes):
            for other_attr in self.attributes[:i] + self.attributes[i + 1 :]:
                if (
                    attr.object_id
                    and other_attr.object_id
                    and attr.object_id == other_attr.object_id
                ):
                    raise ValueError(f"object_id {attr.object_id} is not unique")
                if attr.object_id and attr.object_id == other_attr.name:
                    raise ValueError(f"object_id {attr.object_id} is not unique")
        return self

    def get_attribute(
        self, attribute_name: str
    ) -> Union[
        DeviceAttribute, LazyDeviceAttribute, StaticDeviceAttribute, DeviceCommand
    ]:
        """

        Args:
            attribute_name:

        Returns:

        """
        for attribute in itertools.chain(
            self.attributes,
            self.lazy,
            self.static_attributes,
            self.internal_attributes,
            self.commands,
        ):
            if attribute.name == attribute_name:
                return attribute
        msg = (
            f"Device: {self.device_id}: Could not "
            f"find attribute with name {attribute_name}"
        )
        logger.error(msg)
        raise KeyError(msg)

    def add_attribute(
        self,
        attribute: Union[
            DeviceAttribute, LazyDeviceAttribute, StaticDeviceAttribute, DeviceCommand
        ],
        update: bool = False,
    ) -> None:
        """

        Args:
            attribute:
            update (bool): If 'True' and attribute does already exists tries
                to  update the attribute if not
        Returns:
            None
        """
        try:
            if type(attribute) == DeviceAttribute:
                if attribute.model_dump(exclude_none=True) in [
                    attr.model_dump(exclude_none=True) for attr in self.attributes
                ]:
                    raise ValueError

                self.attributes.append(attribute)
                self.__setattr__(name="attributes", value=self.attributes)
            elif type(attribute) == LazyDeviceAttribute:
                if attribute in self.lazy:
                    raise ValueError

                self.lazy.append(attribute)
                self.__setattr__(name="lazy", value=self.lazy)
            elif type(attribute) == StaticDeviceAttribute:
                if attribute in self.static_attributes:
                    raise ValueError

                self.static_attributes.append(attribute)
                self.__setattr__(name="static_attributes", value=self.static_attributes)
            elif type(attribute) == DeviceCommand:
                if attribute in self.commands:
                    raise ValueError

                self.commands.append(attribute)
                self.__setattr__(name="commands", value=self.commands)
            else:
                raise ValueError
        except ValueError:
            if update:
                self.update_attribute(attribute, append=False)
                logger.warning(
                    "Device: %s: Attribute already " "exists. Will update: \n %s",
                    self.device_id,
                    attribute.model_dump_json(indent=2),
                )
            else:
                logger.error(
                    "Device: %s: Attribute already " "exists: \n %s",
                    self.device_id,
                    attribute.model_dump_json(indent=2),
                )
                raise

    def update_attribute(
        self,
        attribute: Union[
            DeviceAttribute, LazyDeviceAttribute, StaticDeviceAttribute, DeviceCommand
        ],
        append: bool = False,
    ) -> None:
        """
        Updates existing device attribute

        Args:
            attribute: Attribute to add to device configuration
            append (bool): Adds attribute if not exist

        Returns:
            None
        """
        try:
            if type(attribute) == DeviceAttribute:
                idx = self.attributes.index(attribute)
                self.attributes[idx].model_dump().update(attribute.model_dump())
            elif type(attribute) == LazyDeviceAttribute:
                idx = self.lazy.index(attribute)
                self.lazy[idx].model_dump().update(attribute.model_dump())
            elif type(attribute) == StaticDeviceAttribute:
                idx = self.static_attributes.index(attribute)
                self.static_attributes[idx].model_dump().update(attribute.model_dump())
            elif type(attribute) == DeviceCommand:
                idx = self.commands.index(attribute)
                self.commands[idx].model_dump().update(attribute.model_dump())
        except ValueError:
            if append:
                logger.warning(
                    "Device: %s: Could not find " "attribute: \n %s",
                    self.device_id,
                    attribute.model_dump_json(indent=2),
                )
                self.add_attribute(attribute=attribute)
            else:
                msg = (
                    f"Device: {self.device_id}: Could not find "
                    f"attribute: \n {attribute.model_dump_json(indent=2)}"
                )
                raise KeyError(msg)

    def delete_attribute(
        self,
        attribute: Union[
            DeviceAttribute, LazyDeviceAttribute, StaticDeviceAttribute, DeviceCommand
        ],
    ):
        """
        Deletes attribute from device
        Args:
            attribute: ()

        Returns:

        """
        try:
            if type(attribute) == DeviceAttribute:
                self.attributes.remove(attribute)
            elif type(attribute) == LazyDeviceAttribute:
                self.lazy.remove(attribute)
            elif type(attribute) == StaticDeviceAttribute:
                self.static_attributes.remove(attribute)
            elif type(attribute) == DeviceCommand:
                self.commands.remove(attribute)
            else:
                raise ValueError
        except ValueError:
            logger.warning(
                "Device: %s: Could not delete " "attribute: \n %s",
                self.device_id,
                attribute.model_dump_json(indent=2),
            )
            raise

        logger.info(
            "Device: %s: Attribute deleted! \n %s",
            self.device_id,
            attribute.model_dump_json(indent=2),
        )

    def get_command(self, command_name: str):
        """
        Short for self.get_attributes
        Args:
            command_name (str):
        Returns:

        """
        return self.get_attribute(attribute_name=command_name)

    def add_command(self, command: DeviceCommand, update: bool = False):
        """
        Short for self.add_attribute
        Args:
            command (DeviceCommand):
            update (bool): Update command if it already exists
        Returns:
        """
        self.add_attribute(attribute=command, update=update)

    def update_command(self, command: DeviceCommand, append: bool = False):
        """
        Short for self.update_attribute
        Args:
            command:
            append:
        Returns:
        """
        self.update_attribute(attribute=command, append=append)

    def delete_command(self, command: DeviceCommand):
        """
        Short for self.delete_attribute
        Args:
            command:

        Returns:
            None
        """
        self.delete_attribute(attribute=command)



"""
NGSIv2 models for context broker interaction
"""

import json
from typing import Any, List, Dict, Union, Optional, Set, Tuple

from aenum import Enum
from pydantic import (
    field_validator,
    ConfigDict,
    BaseModel,
    Field,
    model_validator,
    SerializeAsAny,
)
from pydantic_core.core_schema import ValidationInfo
from pydantic.types import OnErrorOmit
from filip.models.ngsi_v2.base import (
    EntityPattern,
    Expression,
    BaseAttribute,
    BaseValueAttribute,
    BaseNameAttribute,
)
from filip.models.base import DataType
from filip.utils.validators import (
    validate_fiware_datatype_standard,
    validate_fiware_datatype_string_protect,
    validate_fiware_attribute_value_regex,
    validate_fiware_attribute_name_regex,
)


class GetEntitiesOptions(str, Enum):
    """Options for queries"""

    _init_ = "value __doc__"

    NORMALIZED = "normalized", "Normalized message representation"
    KEY_VALUES = (
        "keyValues",
        "Key value message representation."
        "This mode represents the entity "
        "attributes by their values only, leaving out "
        "the information about type and metadata. "
        "See example "
        "below."
        "Example: "
        "{"
        "  'id': 'R12345',"
        "  'type': 'Room',"
        "  'temperature': 22"
        "}",
    )
    VALUES = (
        "values",
        "Key value message representation. "
        "This mode represents the entity as an array of "
        "attribute values. Information about id and type is "
        "left out. See example below. The order of the "
        "attributes in the array is specified by the attrs "
        "URI param (e.g. attrs=branch,colour,engine). "
        "If attrs is not used, the order is arbitrary. "
        "Example:"
        "[ 'Ford', 'black', 78.3 ]",
    )
    UNIQUE = (
        "unique",
        "unique mode. This mode is just like values mode, "
        "except that values are not repeated",
    )


class PropertyFormat(str, Enum):
    """
    Format to decide if properties of ContextEntity class are returned as
    List of NamedContextAttributes or as Dict of ContextAttributes.
    """

    LIST = "list"
    DICT = "dict"


class ContextAttribute(BaseAttribute, BaseValueAttribute):
    """
    Model for an attribute is represented by a JSON object with the following
    syntax:

    The attribute value is specified by the value property, whose value may
    be any JSON datatype.

    The attribute NGSI type is specified by the type property, whose value
    is a string containing the NGSI type.

    The attribute metadata is specified by the metadata property. Its value
    is another JSON object which contains a property per metadata element
    defined (the name of the property is the name of the metadata element).
    Each metadata element, in turn, is represented by a JSON object
    containing the following properties:

    Values of entity attributes. For adding it you need to nest it into a
    dict in order to give it a name.

    Example:

        >>> data = {"value": <...>,
                    "type": <...>,
                    "metadata": <...>}
        >>> attr = ContextAttribute(**data)

    """

    # although `type` is a required field in the NGSIv2 specification, it is
    # set to optional here to allow for the possibility of setting
    # default-types in child classes. Pydantic will raise the correct error
    # and also exports the correct json-schema.
    def __init__(self, type: str = None, **data):
        if type is None and self.model_fields["type"].default:
            type = self.model_fields["type"].default
        super().__init__(type=type, **data)


class NamedContextAttribute(ContextAttribute, BaseNameAttribute):
    """
    Context attributes are properties of context entities. For example, the
    current speed of a car could be modeled as attribute current_speed of entity
    car-104.

    In the NGSI data model, attributes have an attribute name, an attribute type
    an attribute value and metadata.
    """

    pass


class ContextEntityKeyValues(BaseModel):
    """
    Base Model for an entity is represented by a JSON object with the following
    syntax.

    The entity id is specified by the object's id property, whose value
    is a string containing the entity id.

    The entity type is specified by the object's type property, whose value
    is a string containing the entity's type name.

    """

    model_config = ConfigDict(
        extra="allow", validate_default=True, validate_assignment=True
    )
    id: str = Field(
        ...,
        title="Entity Id",
        description="Id of an entity in an NGSI context broker. Allowed "
        "characters are the ones in the plain ASCII set, except "
        "the following ones: control characters, "
        "whitespace, &, ?, / and #.",
        json_schema_extra={"example": "Bcn-Welt"},
        max_length=256,
        min_length=1,
        frozen=True,
    )
    valid_id = field_validator("id")(validate_fiware_datatype_standard)
    type: Union[str, Enum] = Field(
        ...,
        title="Entity Type",
        description="Id of an entity in an NGSI context broker. "
        "Allowed characters are the ones in the plain ASCII set, "
        "except the following ones: control characters, "
        "whitespace, &, ?, / and #.",
        json_schema_extra={"example": "Room"},
        max_length=256,
        min_length=1,
        frozen=True,
    )
    valid_type = field_validator("type")(validate_fiware_datatype_standard)

    # although `type` is a required field in the NGSIv2 specification, it is
    # set to optional here to allow for the possibility of setting
    # default-types in child classes. Pydantic will raise the correct error
    # and also exports the correct json-schema.
    def __init__(self, id: str, type: Union[str, Enum] = None, **data):
        # this allows to set the type of the entity in child classes
        if type is None:
            if isinstance(self.model_fields["type"].default, str):
                type = self.model_fields["type"].default
            else:
                # if this statement is reached not proper default-value for
                # `type` was found and pydantic will raise the correct error
                super().__init__(id=id, **data)
        # This will result in usual behavior
        data.update(self._validate_attributes(data))
        super().__init__(id=id, type=type, **data)

    # Validation of attributes
    @classmethod
    def _validate_attributes(cls, data: dict):
        """
        Validate attribute name and value of the entity in keyvalues format
        """
        for attr_name, attr_value in data.items():
            if isinstance(attr_value, str):
                validate_fiware_attribute_value_regex(attr_value)
            validate_fiware_attribute_name_regex(attr_name)
        return data

    def get_attributes(self) -> dict:
        """
        Get the attribute of the entity with the given name in
        dict format

        Returns:
            dict
        """
        return self.model_dump(exclude={"id", "type"})

    def to_normalized(self):
        attrs = []
        for key, value in self.get_attributes().items():
            attr_type = (
                DataType.NUMBER.value
                if isinstance(value, int) or isinstance(value, float)
                else (
                    DataType.TEXT.value
                    if isinstance(value, str)
                    else DataType.OBJECT.value
                )
            )
            attr = NamedContextAttribute(name=key, value=value, type=attr_type)
            attrs.append(attr)
        entity = ContextEntity(self.id, self.type)
        entity.add_attributes(attrs)
        return entity


class ContextEntity(ContextEntityKeyValues):
    """
    Context entities, or simply entities, are the center of gravity in the
    FIWARE NGSI information model. An entity represents a thing, i.e., any
    physical or logical object (e.g., a sensor, a person, a room, an issue in
    a ticketing system, etc.). Each entity has an entity id.
    Furthermore, the type system of FIWARE NGSI enables entities to have an
    entity type. Entity types are semantic types; they are intended to describe
    the type of thing represented by the entity. For example, a context
    entity #with id sensor-365 could have the type temperatureSensor.

    Each entity is uniquely identified by the combination of its id and type.

    The entity id is specified by the object's id property, whose value
    is a string containing the entity id.

    The entity type is specified by the object's type property, whose value
    is a string containing the entity's type name.

    Entity attributes are specified by additional properties, whose names are
    the name of the attribute and whose representation is described by the
    "ContextAttribute"-model. Obviously, `id` and `type` are
    not allowed as attribute names.

    Example::

        >>> data = {'id': 'MyId',
                    'type': 'MyType',
                    'my_attr': {'value': 20, 'type': 'Number'}}

        >>> entity = ContextEntity(**data)

    """

    model_config = ConfigDict(
        extra="allow", validate_default=True, validate_assignment=True
    )

    # although `type` is a required field in the NGSIv2 specification, it is
    # set to optional here to allow for the possibility of setting
    # default-types in child classes. Pydantic will raise the correct error
    # and also exports the correct json-schema.
    def __init__(self, id: str, type: str = None, **data):
        # There is currently no validation for extra fields
        data.update(self._validate_attributes(data))
        # case where type is None to raise correct error message
        if type is None:
            super().__init__(id=id, **data)
        else:
            super().__init__(id=id, type=type, **data)

    # Validation of attributes
    @classmethod
    def _validate_attributes(cls, data: dict):
        """
        Validate attributes of the entity if the attribute is not a model
        field and the type is not already a subtype of ContextAttribute
        """
        attrs = {
            key: ContextAttribute.model_validate(attr)
            for key, attr in data.items()
            if (
                # validate_fiware_attribute_value_regex(key) not in cls.model_fields
                validate_fiware_attribute_name_regex(key) not in cls.model_fields
                and not isinstance(attr, ContextAttribute)
                # key not in cls.model_fields
                # and not isinstance(attr, ContextAttribute)
            )
        }

        return attrs

    @field_validator("*")
    @classmethod
    def check_attributes(cls, value, info: ValidationInfo):
        """
        Check whether all model fields are of subtype of ContextAttribute to
        ensure full functionality.
        """
        if info.field_name in ["id", "type"]:
            return value

        if info.field_name in cls.model_fields:
            if not (
                isinstance(value, ContextAttribute)
                or value == cls.model_fields[info.field_name].default
            ):
                raise ValueError(
                    f"Attribute {info.field_name} must be a of "
                    f"type or subtype ContextAttribute"
                )
        return value

    @model_validator(mode="after")
    @classmethod
    def check_attributes_after(cls, values):
        try:
            for attr in values.model_extra:
                if not isinstance(values.__getattr__(attr), ContextAttribute):
                    raise ValueError(
                        f"Attribute {attr} must be a of type or "
                        f"subtype ContextAttribute. You most "
                        f"likely tried to directly assign an "
                        f"attribute without converting it to a "
                        f"proper Attribute-Type!"
                    )
        except TypeError:
            pass
        return values

    # API for attributes and commands
    def add_attributes(
        self, attrs: Union[Dict[str, ContextAttribute], List[NamedContextAttribute]]
    ) -> None:
        """
        Add attributes (properties, relationships) to entity

        Args:
            attrs: Dict[str, ContextAttribute]: {NAME for attr : Attribute} or
                   List[NamedContextAttribute]

        Returns:
            None
        """
        if isinstance(attrs, list):
            attrs = {
                attr.name: ContextAttribute(**attr.model_dump(exclude={"name"}))
                for attr in attrs
            }
        for key, attr in attrs.items():
            self.__setattr__(name=key, value=attr)

    def get_attributes(
        self,
        whitelisted_attribute_types: Optional[List[DataType]] = None,
        blacklisted_attribute_types: Optional[List[DataType]] = None,
        response_format: Union[str, PropertyFormat] = PropertyFormat.LIST,
        strict_data_type: bool = True,
    ) -> Union[List[NamedContextAttribute], Dict[str, ContextAttribute]]:
        """
        Get attributes or a subset from the entity.

        Args:
            whitelisted_attribute_types: Optional list, if given only
                attributes matching one of the types are returned
            blacklisted_attribute_types: Optional list, if given all
                attributes are returned that do not match a list entry
            response_format: Wanted result format,
                                List -> list of NamedContextAttributes
                                Dict -> dict of {name: ContextAttribute}
            strict_data_type: whether to restrict the data type to pre-defined
                types, True by default.
                True  -> Only return the attributes with pre-defined types,
                False -> Do not restrict the data type.
        Raises:
            AssertionError, if both a white and a black list is given
        Returns:
            List[NamedContextAttribute] or Dict[str, ContextAttribute]
        """

        response_format = PropertyFormat(response_format)

        assert (
            whitelisted_attribute_types is None or blacklisted_attribute_types is None
        ), "Only whitelist or blacklist is allowed"

        if whitelisted_attribute_types is not None:
            attribute_types = whitelisted_attribute_types
        elif blacklisted_attribute_types is not None:
            attribute_types = [
                att_type
                for att_type in list(DataType)
                if att_type not in blacklisted_attribute_types
            ]
        else:
            attribute_types = [att_type for att_type in list(DataType)]

        if response_format == PropertyFormat.DICT:
            if strict_data_type:
                return {
                    key: ContextAttribute(**value)
                    for key, value in self.model_dump().items()
                    if key not in ContextEntity.model_fields
                    and value.get("type") in [att.value for att in attribute_types]
                }
            else:
                return {
                    key: ContextAttribute(**value)
                    for key, value in self.model_dump().items()
                    if key not in ContextEntity.model_fields
                }
        else:
            if strict_data_type:
                return [
                    NamedContextAttribute(name=key, **value)
                    for key, value in self.model_dump().items()
                    if key not in ContextEntity.model_fields
                    and value.get("type") in [att.value for att in attribute_types]
                ]
            else:
                return [
                    NamedContextAttribute(name=key, **value)
                    for key, value in self.model_dump().items()
                    if key not in ContextEntity.model_fields
                ]

    def update_attribute(
        self, attrs: Union[Dict[str, ContextAttribute], List[NamedContextAttribute]]
    ) -> None:
        """
        Update attributes of an entity. Overwrite the current held value
        for the attribute with the value contained in the corresponding given
        attribute

        Args:
            attrs: List of NamedContextAttributes,
                   Dict of {attribute_name: ContextAttribute}
        Raises:
            NameError, if the attribute does not currently exist in the entity
        Returns:
            None
        """
        if isinstance(attrs, list):
            attrs = {
                attr.name: ContextAttribute(**attr.model_dump(exclude={"name"}))
                for attr in attrs
            }

        existing_attribute_names = self.get_attribute_names()
        for key, attr in attrs.items():
            if key not in existing_attribute_names:
                raise NameError
            self.__setattr__(name=key, value=attr)

    def get_attribute_names(self) -> Set[str]:
        """
        Returns a set with all attribute names of this entity

        Returns:
            Set[str]
        """

        return {
            key for key in self.model_dump() if key not in ContextEntity.model_fields
        }

    def delete_attributes(
        self,
        attrs: Union[
            Dict[str, ContextAttribute], List[NamedContextAttribute], List[str]
        ],
    ):
        """
        Delete the given attributes from the entity

        Args:
            attrs:  - Dict {name: ContextAttribute}
                    - List[NamedContextAttribute]
                    - List[str] -> names of attributes
        Raises:
            Exception: if one of the given attrs does not represent an
                       existing argument
        """

        names: List[str] = []
        if isinstance(attrs, list):
            for entry in attrs:
                if isinstance(entry, str):
                    names.append(entry)
                elif isinstance(entry, NamedContextAttribute):
                    names.append(entry.name)
        else:
            names.extend(list(attrs.keys()))
        for name in names:
            delattr(self, name)

    def get_attribute(self, attribute_name: str) -> NamedContextAttribute:
        """
        Get the attribute of the entity with the given name

        Args:
            attribute_name (str): Name of attribute

        Raises:
            KeyError, if no attribute with given name exists

        Returns:
            NamedContextAttribute
        """
        for attr in self.get_attributes():
            if attr.name == attribute_name:
                return attr
        raise KeyError(f"Attribute '{attribute_name}' not in entity")

    def get_properties(
        self, response_format: Union[str, PropertyFormat] = PropertyFormat.LIST
    ) -> Union[List[NamedContextAttribute], Dict[str, ContextAttribute]]:
        """
        Returns all attributes of the entity that are not of type Relationship,
        and are not auto generated command attributes

        Args:
            response_format: Wanted result format,
                                List -> list of NamedContextAttributes
                                Dict -> dict of {name: ContextAttribute}

        Returns:
            [NamedContextAttribute] or {name: ContextAttribute}
        """
        pre_filtered_attrs = self.get_attributes(
            blacklisted_attribute_types=[DataType.RELATIONSHIP],
            response_format=PropertyFormat.LIST,
        )

        all_command_attributes_names = set()
        for command in self.get_commands():
            (c, c_status, c_info) = self.get_command_triple(command.name)
            all_command_attributes_names.update([c.name, c_status.name, c_info.name])

        property_attributes = []
        for attr in pre_filtered_attrs:
            if attr.name not in all_command_attributes_names:
                property_attributes.append(attr)

        if response_format == PropertyFormat.LIST:
            return property_attributes
        else:
            return {
                p.name: ContextAttribute(**p.model_dump(exclude={"name"}))
                for p in property_attributes
            }

    def get_relationships(
        self, response_format: Union[str, PropertyFormat] = PropertyFormat.LIST
    ) -> Union[List[NamedContextAttribute], Dict[str, ContextAttribute]]:
        """
        Get all relationships of the context entity

        Args:
            response_format: Wanted result format,
                                List -> list of NamedContextAttributes
                                Dict -> dict of {name: ContextAttribute}

        Returns:
            [NamedContextAttribute] or {name: ContextAttribute}

        """
        return self.get_attributes(
            whitelisted_attribute_types=[DataType.RELATIONSHIP],
            response_format=response_format,
        )

    def get_commands(
        self, response_format: Union[str, PropertyFormat] = PropertyFormat.LIST
    ) -> Union[List[NamedContextAttribute], Dict[str, ContextAttribute]]:
        """
        Get all commands of the context entity. Only works if the commands
        were autogenerated by Fiware from an Device.

        Args:
            response_format: Wanted result format,
                                List -> list of NamedContextAttributes
                                Dict -> dict of {name: ContextAttribute}

        Returns:
            [NamedContextAttribute] or {name: ContextAttribute}
        """

        # if an attribute with name n is a command, its type does not need to
        # be COMMAND.
        # But the attributes name_info (type: commandResult) and
        # name_status(type: commandStatus) need to exist. (Autogenerated)

        # Search all attributes of type commandStatus, check for each if a
        # corresponding _info exists and if also a fitting attribute exists
        # we know: that is a command.

        commands = []
        for status_attribute in self.get_attributes(
            whitelisted_attribute_types=[DataType.COMMAND_STATUS]
        ):
            if not status_attribute.name.split("_")[-1] == "status":
                continue
            base_name = status_attribute.name[:-7]

            try:
                info_attribute = self.get_attribute(f"{base_name}_info")
                if not info_attribute.type == DataType.COMMAND_RESULT:
                    continue

                attribute = self.get_attribute(base_name)
                commands.append(attribute)
            except KeyError:
                continue

        if response_format == PropertyFormat.LIST:
            return commands
        else:
            return {
                cmd.name: ContextAttribute(**cmd.model_dump(exclude={"name"}))
                for cmd in commands
            }

    def get_command_triple(
        self, command_attribute_name: str
    ) -> Tuple[NamedContextAttribute, NamedContextAttribute, NamedContextAttribute]:
        """
        Returns for a given command attribute name all three corresponding
        attributes as triple

        Args:
            command_attribute_name: Name of the command attribute

        Raises:
            KeyError, if the given name does not belong to a command attribute

        Returns:
            (Command, Command_status, Command_info)
        """

        commands = self.get_commands(response_format=PropertyFormat.DICT)

        if command_attribute_name not in commands:
            raise KeyError(f"Command '{command_attribute_name}' not in commands")

        command = self.get_attribute(command_attribute_name)

        # as the given name was found as a valid command, we know that the
        # status and info attributes exist correctly
        command_status = self.get_attribute(f"{command_attribute_name}_status")
        command_info = self.get_attribute(f"{command_attribute_name}_info")

        return command, command_status, command_info

    def to_keyvalues(self):
        attrs = {
            attr: value.value
            for attr, value in self.get_attributes(
                response_format=PropertyFormat.DICT
            ).items()
        }
        entity = ContextEntityKeyValues(self.id, self.type, **attrs)
        return entity

    def to_normalized(self):
        raise AttributeError("This method is not available in ContextEntity")


class ContextEntityList(BaseModel):
    """
    Collection model for a list of context entities
    """

    entities: List[OnErrorOmit[ContextEntity]]


class ContextEntityKeyValuesList(BaseModel):
    """
    Collection model for a list of context entities in key-values format
    """

    entities: List[OnErrorOmit[ContextEntityKeyValues]]


class ContextEntityValidationList(ContextEntityList):
    """
    Collection model for a list of valid and invalid context entities
    """

    invalid_entities: List[str]


class ContextEntityKeyValuesValidationList(ContextEntityKeyValuesList):
    """
    Collection model for a list of valid and invalid context entities in key-values format
    """

    invalid_entities: List[str]


class Query(BaseModel):
    """
    Model for queries
    """

    entities: List[EntityPattern] = Field(
        description="a list of entities to search for. Each element is "
        "represented by a JSON object"
    )
    attrs: Optional[List[str]] = Field(
        default=None,
        description="List of attributes to be provided "
        "(if not specified, all attributes).",
    )
    expression: Optional[Expression] = Field(
        default=None,
        description="An expression composed of q, mq, georel, geometry and " "coords",
    )
    metadata: Optional[List[str]] = Field(
        default=None,
        description="a list of metadata names to include in the response. "
        'See "Filtering out attributes and metadata" section for '
        "more detail.",
    )


class ActionType(str, Enum):
    """
    Options for queries
    """

    _init_ = "value __doc__"
    APPEND = (
        "append",
        "maps to POST /v2/entities (if the entity does not "
        "already exist) or POST /v2/entities/<id>/attrs (if "
        "the entity already exists). ",
    )
    APPEND_STRICT = (
        "appendStrict",
        "maps to POST /v2/entities (if the "
        "entity does not already exist) or POST "
        "/v2/entities/<id>/attrs?options=append "
        "(if the entity already exists).",
    )
    UPDATE = "update", "maps to PATCH /v2/entities/<id>/attrs."
    DELETE = (
        "delete",
        "maps to DELETE /v2/entities/<id>/attrs/<attrName> on "
        "every attribute included in the entity or to DELETE "
        "/v2/entities/<id> if no attribute were included in "
        "the entity.",
    )
    REPLACE = "replace", "maps to PUT /v2/entities/<id>/attrs"


class Update(BaseModel):
    """
    Model for update action
    """

    action_type: Union[ActionType, str] = Field(
        alias="actionType",
        description="actionType, to specify the kind of update action to do: "
        "either append, appendStrict, update, delete, or replace. ",
    )
    entities: SerializeAsAny[List[Union[ContextEntity, ContextEntityKeyValues]]] = (
        Field(
            description="an array of entities, each entity specified using the "
            "JSON entity representation format "
        )
    )

    @field_validator("action_type")
    @classmethod
    def check_action_type(cls, action):
        """
        validates action_type
        Args:
            action: field action_type
        Returns:
            action_type
        """
        return ActionType(action)


class Command(BaseModel):
    """
    Class for sending commands to IoT Devices.
    Note that the command must be registered via an IoT-Agent. Internally
    FIWARE uses its registration mechanism in order to connect the command
    with an IoT-Device
    """

    type: DataType = Field(
        default=DataType.COMMAND,
        description="Command must have the type command",
        # const=True
    )
    value: Any = Field(
        description="Any json serializable command that will "
        "be forwarded to the connected IoT device"
    )

    @field_validator("value")
    @classmethod
    def check_value(cls, value):
        """
        Check if value is json serializable
        Args:
            value: value field
        Returns:
            value
        """
        try:
            json.dumps(value)
        except:
            raise ValueError(f"Command value {value} " f"is not serializable")
        return value


class NamedCommand(Command):
    """
    Class for sending command to IoT-Device.
    Extend :class: Command with command Name
    """

    name: str = Field(
        description="Name of the command",
        max_length=256,
        min_length=1,
    )
    valid_name = field_validator("name")(validate_fiware_attribute_name_regex)



from datetime import datetime, timezone


def transform_to_utc_datetime(dt: datetime) -> datetime:
    """
    Converts datetime object to utc datetime object with zone

    Args:
        dt:

    Returns:

    """
    return dt.astimezone(tz=timezone.utc)


def convert_datetime_to_iso_8601_with_z_suffix(dt: datetime) -> str:
    """
    Converts datetime object to iso8601 notation with z-suffix

    Args:
        dt: datetime object

    Returns:
        String in iso 8601 notation with z-suffix
    """
    dt = transform_to_utc_datetime(dt)
    return dt.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"



"""Setup.py script for the FiLiP-Library"""

import re
import setuptools

# read the contents of your README file
from pathlib import Path

readme_path = Path(__file__).parent.joinpath("README.md")
LONG_DESCRIPTION = readme_path.read_text(encoding="utf-8")

INSTALL_REQUIRES = [
    "aenum~=3.1.15",
    "datamodel_code_generator[http]~=0.25.0",
    "paho-mqtt~=2.0.0",
    "pandas_datapackage_reader~=0.18.0",
    "pydantic>=2.6.0,<2.9.0",
    "pydantic-settings>=2.0.0,<2.3.0",
    "geojson_pydantic~=1.0.2",
    "stringcase>=1.2.0",
    "regex~=2023.10.3",
    "requests~=2.32.0",
    "rapidfuzz~=3.4.0",
    "geojson-pydantic~=1.0.2",
    "wget~=3.2",
    "PyLD~=2.0.4",
    "pyjexl~=0.3.0",
    "packaging~=24.1",
]

SETUP_REQUIRES = INSTALL_REQUIRES.copy()


def get_version():
    init_py = Path(__file__).parent.joinpath("filip", "__init__.py")
    with open(init_py, "r") as f:
        content = f.read()
    # Using a regular expression to find the version string
    match = re.search(r"^__version__\s*=\s*['\"]([^'\"]*)['\"]", content, re.M)
    if match:
        return match.group(1)
    raise RuntimeError("Unable to find version string.")


VERSION = get_version()

setuptools.setup(
    name="filip",
    version=VERSION,
    author="RWTH Aachen University, E.ON Energy Research Center, Institute\
        of Energy Efficient Buildings and Indoor Climate",
    author_email="ebc-tools@eonerc.rwth-aachen.de",
    description="[FI]WARE [Li]brary for [P]ython",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url="https://github.com/RWTH-EBC/filip",
    download_url=f"https://github.com/RWTH-EBC/FiLiP/archive/refs/tags/v{VERSION}.tar.gz",
    project_urls={
        "Documentation": "https://rwth-ebc.github.io/FiLiP/master/docs/index.html",
        "Source": "https://github.com/RWTH-EBC/filip",
        "Download": f"https://github.com/RWTH-EBC/FiLiP/archive/refs/tags/v{VERSION}.tar.gz",
    },
    # Specify the Python versions you support here. In particular, ensure
    # that you indicate whether you support Python 2, Python 3 or both.
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Scientific/Engineering",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: BSD License",
    ],
    keywords=["iot", "fiware", "semantic"],
    packages=setuptools.find_packages(
        exclude=["tests", "tests.*", "img", "tutorials.*", "tutorials"]
    ),
    package_data={"filip": ["data/unece-units/*.csv"]},
    setup_requires=SETUP_REQUIRES,
    # optional modules
    extras_require={
        "development": ["pre-commit~=4.0.1"],
        "semantics": ["igraph~=0.11.2", "rdflib>=6.0.0,<=6.1.1"],
        "tutorials": ["plotly==5.24.1", "matplotlib~=3.4.3"],
        ":python_version < '3.9'": ["pandas~=1.3.5"],
        ":python_version >= '3.9'": ["pandas>=2.1.4,<2.4.0"],
    },
    install_requires=INSTALL_REQUIRES,
    python_requires=">=3.8",
)



"""
This module contains NGSIv2 models for context registrations in the context
broker.
"""

from typing import List, Union, Optional
from datetime import datetime
from aenum import Enum
from pydantic import ConfigDict, BaseModel, Field
from filip.models.ngsi_v2.base import EntityPattern, Expression, Http, Status


class ForwardingMode(str, Enum):
    _init_ = "value __doc__"

    NONE = "none", "This provider does not support request forwarding."
    QUERY = "query", "This provider only supports request forwarding to " "query data."
    UPDATE = (
        "update",
        "This provider only supports request forwarding to " "update data.",
    )
    ALL = (
        "all",
        "This provider supports both query and update forwarding "
        "requests. (Default value)",
    )


class Provider(BaseModel):
    http: Http = Field(
        description="It is used to convey parameters for providers that "
        "deliver information through the HTTP protocol. (Only "
        "protocol supported nowadays). It must contain a subfield "
        "named url with the URL that serves as the endpoint that "
        "offers the providing interface. The endpoint must not "
        "include the protocol specific part (for instance "
        "/v2/entities). "
    )
    supportedForwardingMode: ForwardingMode = Field(
        default=ForwardingMode.ALL,
        description="It is used to convey the forwarding mode supported by "
        "this context provider. By default all.",
    )


class ForwardingInformation(BaseModel):
    model_config = ConfigDict(frozen=True)

    timesSent: int = Field(
        description="(not editable, only present in GET operations): "
        "Number of forwarding requests sent due to this "
        "registration."
    )
    lastForwarding: datetime = Field(
        description="(not editable, only present in GET operations): "
        "Last forwarding timestamp in ISO8601 format."
    )
    lastFailure: Optional[datetime] = Field(
        default=None,
        description="(not editable, only present in GET operations): "
        "Last failure timestamp in ISO8601 format. Not present "
        "if registration has never had a problem with forwarding.",
    )
    lastSuccess: Optional[datetime] = Field(
        default=None,
        description="(not editable, only present in GET operations): "
        "Timestamp in ISO8601 format for last successful "
        "request forwarding. Not present if registration has "
        "never had a successful notification.",
    )


class DataProvided(BaseModel):
    """
    Model for provided data
    """

    entities: List[EntityPattern] = Field(
        description="A list of objects, each one composed by an entity object"
    )
    attrs: Optional[List[str]] = Field(
        default=None,
        description="List of attributes to be provided "
        "(if not specified, all attributes)",
    )
    expression: Optional[Union[str, Expression]] = Field(
        default=None,
        description="By means of a filtering expression, allows to express "
        "what is the scope of the data provided. Currently only "
        "geographical scopes are supported ",
    )


class Registration(BaseModel):
    """
    A Context Registration allows to bind external context information
    sources so that they can play the role of providers of certain subsets
    (entities, attributes) of the context information space, including those
    located at specific geographical areas.
    """

    id: Optional[str] = Field(
        default=None,
        description="Unique identifier assigned to the registration. "
        "Automatically generated at creation time.",
    )
    description: Optional[str] = Field(
        default=None,
        description="A free text used by the client to describe the " "registration.",
        json_schema_extra={"example": "Relative Humidity Context Source"},
    )
    provider: Provider = Field(
        description="Object that describes the context source registered.",
        json_schema_extra={"example": '"http": {"url": "http://localhost:1234"}'},
    )
    dataProvided: DataProvided = Field(
        description="Object that describes the data provided by this source",
        json_schema_extra={
            "example": "{"
            '   "entities": [{"id": "room2", "type": "Room"}],'
            '   "attrs": ["relativeHumidity"]'
            "},"
        },
    )
    status: Optional[Status] = Field(
        default=Status.ACTIVE,
        description="Either active (for active registration) or inactive "
        "(for inactive registration). If this field is not "
        "provided at registration creation time, new registration "
        "are created with the active status, which can be changed"
        " by clients afterwards. For expired registration, this "
        "attribute is set to expired (no matter if the client "
        "updates it to active/inactive). Also, for subscriptions "
        "experiencing problems with notifications, the status is "
        "set to failed. As soon as the notifications start working "
        "again, the status is changed back to active.",
    )
    expires: Optional[datetime] = Field(
        default=None,
        description="Registration expiration date in ISO8601 format. "
        "Permanent registrations must omit this field.",
    )
    forwardingInformation: Optional[ForwardingInformation] = Field(
        default=None,
        description="Information related to the forwarding operations made "
        "against the provider. Automatically provided by the "
        "implementation, in the case such implementation supports "
        "forwarding capabilities.",
    )



"""
Utility module for helper functions
"""

from .validators import validate_http_url, validate_mqtt_url
from .datetime import (
    convert_datetime_to_iso_8601_with_z_suffix,
    transform_to_utc_datetime,
)



"""
Helper functions to handle the devices related with the IoT Agent
"""

import warnings
from typing import List, Union
from filip.models.ngsi_v2.iot import Device
from filip.utils.filter import filter_device_list as filter_device_list_new


def filter_device_list(
    devices: List[Device],
    device_ids: Union[str, List[str]] = None,
    entity_names: Union[str, List[str]] = None,
    entity_types: Union[str, List[str]] = None,
) -> List[Device]:
    """
    Filter the given device list based on conditions

    Args:
        devices: device list that need to be filtered
        device_ids: A list of device_id as filter condition
        entity_names: A list of entity_name (e.g. entity_id) as filter condition.
        entity_types: A list of entity_type as filter condition

    Returns:
        List of matching devices
    """
    warnings.warn(
        "This function has been moved to 'filip.utils.filter' "
        "and will be removed from this module in future releases!",
        DeprecationWarning,
    )

    return filter_device_list_new(
        devices=devices,
        device_ids=device_ids,
        entity_names=entity_names,
        entity_types=entity_types,
    )



from typing import List, Optional, Literal
from pydantic import (
    ConfigDict,
    BaseModel,
    Field,
    HttpUrl,
    AnyUrl,
    field_validator,
    model_validator,
)
import dateutil.parser
from filip.models.ngsi_ld.base import GeoQuery, validate_ngsi_ld_query


class EntityInfo(BaseModel):
    """
    In v1.3.1 it is specified as EntityInfo
    In v1.6.1 it is specified in a new data type, namely EntitySelector
    """

    id: Optional[str] = Field(default=None, description="Entity identifier (valid URI)")
    idPattern: Optional[str] = Field(
        default=None, description="Regular expression as per IEEE POSIX 1003.2™ [11]"
    )
    type: str = Field(
        description="Fully Qualified Name of an Entity Type or the Entity Type Name as a "
        "short-hand string. See clause 4.6.2"
    )
    model_config = ConfigDict(populate_by_name=True)


class KeyValuePair(BaseModel):
    key: str
    value: str


class Endpoint(BaseModel):
    """
    This datatype represents the parameters that are required in order to define
    an endpoint for notifications. This can include the endpoint's URI, a
    generic{key, value} array, named receiverInfo, which contains, in a
    generalized form, whatever extra information the broker shall convey to the
    receiver in order for the broker to successfully communicate with
    receiver (e.g Authorization material), or for the receiver to correctly
    interpret the received content (e.g. the Link URL to fetch an @context).

    Additionally, it can include another generic{key, value} array, named
    notifierInfo, which contains the configuration that the broker needs to
    know in order to correctly set up the communication channel towards the
    receiver

    Example of "receiverInfo"
        "receiverInfo": [
            {
              "key": "H1",
              "value": "123"
            },
            {
              "key": "H2",
              "value": "456"
            }
          ]

    Example of "notifierInfo"
        "notifierInfo": [
            {
              "key": "MQTT-Version",
              "value": "mqtt5.0"
            }
          ]
    """

    uri: AnyUrl = Field(description="Dereferenceable URI")
    accept: Optional[str] = Field(
        default=None,
        description="MIME type for the notification payload body "
        "(application/json, application/ld+json, "
        "application/geo+json)",
    )
    receiverInfo: Optional[List[KeyValuePair]] = Field(
        default=None,
        description="Generic {key, value} array to convey optional information "
        "to the receiver",
    )
    notifierInfo: Optional[List[KeyValuePair]] = Field(
        default=None,
        description="Generic {key, value} array to set up the communication " "channel",
    )
    model_config = ConfigDict(populate_by_name=True)

    @field_validator("uri")
    @classmethod
    def check_uri(cls, uri: AnyUrl):
        if uri.scheme not in ("http", "mqtt"):
            raise ValueError("NGSI-LD currently only support http and mqtt")
        return uri

    @field_validator("notifierInfo")
    @classmethod
    def check_notifier_info(cls, notifierInfo: List[KeyValuePair]):
        # TODO add validation of notifierInfo for MQTT notification
        return notifierInfo


class NotificationParams(BaseModel):
    """
    NGSI-LD Notification model. It contains the parameters that allow to
    convey the details of a notification, as described in NGSI-LD Spec section 5.2.14
    """

    attributes: Optional[List[str]] = Field(
        default=None,
        description="Entity Attribute Names (Properties or Relationships) to be included "
        "in the notification payload body. If undefined, it will mean all Attributes",
    )
    format: Optional[str] = Field(
        default="normalized",
        description="Conveys the representation format of the entities delivered at "
        "notification time. By default, it will be in normalized format",
    )
    endpoint: Endpoint = Field(..., description="Notification endpoint details")
    # status can either be "ok" or "failed"
    status: Literal["ok", "failed"] = Field(
        default="ok",
        description="Status of the Notification. It shall be 'ok' if the last attempt "
        "to notify the subscriber succeeded. It shall be 'failed' if the last"
        " attempt to notify the subscriber failed",
    )

    # Additional members
    timesSent: Optional[int] = Field(
        default=None,
        description="Number of times that the notification was sent. Provided by the "
        "system when querying the details of a subscription",
    )
    lastNotification: Optional[str] = Field(
        default=None,
        description="Timestamp corresponding to the instant when the last notification "
        "was sent. Provided by the system when querying the details of a subscription",
    )
    lastFailure: Optional[str] = Field(
        default=None,
        description="Timestamp corresponding to the instant when the last notification"
        " resulting in failure was sent. Provided by the system when querying the details of a subscription",
    )
    lastSuccess: Optional[str] = Field(
        default=None,
        description="Timestamp corresponding to the instant when the last successful "
        "notification was sent. Provided by the system when querying the details of a subscription",
    )
    model_config = ConfigDict(populate_by_name=True)


class TemporalQuery(BaseModel):
    """
    Temporal query according to NGSI-LD Spec section 5.2.21

    timerel:
        Temporal relationship, one of "before", "after" and "between".
        "before": before the time specified by timeAt.
        "after": after the time specified by timeAt.
        "between": after the time specified by timeAt and before the time specified by
            endtimeAt
    timeAt:
        A DateTime object following ISO 8061, e.g. 2007-12-24T18:21Z
    endTimeAt (optional):
        A DateTime object following ISO 8061, e.g. 2007-12-24T18:21Z
        Only required when timerel="between"
    timeproperty: str
        Representing a Propertyname of the Property that contains the temporal data that
        will be used to resolve the temporal query. If not specified, the default is
        "observedAt"

    """

    model_config = ConfigDict(populate_by_name=True)
    timerel: Literal["before", "after", "between"] = Field(
        ...,
        description="String representing the temporal relationship as defined by clause "
        "4.11 (Allowed values: 'before', 'after', and 'between') ",
    )
    timeAt: str = Field(
        ...,
        description="String representing the timeAt parameter as defined by clause "
        "4.11. It shall be a DateTime ",
    )
    endTimeAt: Optional[str] = Field(
        default=None,
        description="String representing the endTimeAt parameter as defined by clause "
        "4.11. It shall be a DateTime. Cardinality shall be 1 if timerel is "
        "equal to 'between' ",
    )
    timeproperty: Optional[str] = Field(
        default=None,
        description="String representing a Property name. The name of the Property that "
        "contains the temporal data that will be used to resolve the "
        "temporal query. If not specified, ",
    )

    @field_validator("timeAt", "endTimeAt")
    @classmethod
    def check_uri(cls, v: str):
        if not v:
            return v
        else:
            try:
                dateutil.parser.isoparse(v)
            except ValueError:
                raise ValueError("timeAt must be in ISO8061 format")
            return v

    # when timerel=between, endTimeAt must be specified
    @model_validator(mode="after")
    def check_passwords_match(self) -> "TemporalQuery":
        if self.timerel == "between" and self.endTimeAt is None:
            raise ValueError('When timerel="between", endTimeAt must be specified')
        return self


class SubscriptionLD(BaseModel):
    """
    Context Subscription model according to NGSI-LD Spec section 5.2.12
    """

    id: Optional[str] = Field(
        default=None, description="Subscription identifier (JSON-LD @id)"
    )
    type: str = Field(default="Subscription", description="JSON-LD @type")
    subscriptionName: Optional[str] = Field(
        default=None, description="A (short) name given to this Subscription"
    )
    description: Optional[str] = Field(
        default=None, description="Subscription description"
    )
    entities: Optional[List[EntityInfo]] = Field(
        default=None, description="Entities subscribed"
    )
    watchedAttributes: Optional[List[str]] = Field(
        default=None, description="Watched Attributes (Properties or Relationships)"
    )
    notificationTrigger: Optional[List[str]] = Field(
        default=None, description="Notification triggers"
    )
    timeInterval: Optional[int] = Field(
        default=None, description="Time interval in seconds"
    )
    q: Optional[str] = Field(
        default=None,
        description="Query met by subscribed entities to trigger the notification",
    )

    @field_validator("q")
    @classmethod
    def check_q(cls, v: str):
        return validate_ngsi_ld_query(v)

    geoQ: Optional[GeoQuery] = Field(
        default=None,
        description="Geoquery met by subscribed entities to trigger the notification",
    )
    csf: Optional[str] = Field(default=None, description="Context source filter")
    isActive: bool = Field(
        default=True,
        description="Indicates if the Subscription is under operation (True) or paused (False)",
    )
    notification: NotificationParams = Field(..., description="Notification details")
    expiresAt: Optional[str] = Field(
        default=None, description="Expiration date for the subscription"
    )
    throttling: Optional[int] = Field(
        default=None,
        description="Minimal period of time in seconds between two consecutive notifications."
        "It must be greater than 0 if set. If not set, it means no throttling is used.",
    )

    @field_validator("throttling")
    @classmethod
    def check_throttling(cls, throttling):
        """
        Validate the throttling value. It must be greater than 0 if provided.
        """
        if throttling is not None and throttling <= 0:
            raise ValueError("Throttling must be greater than 0")
        return throttling

    temporalQ: Optional[TemporalQuery] = Field(
        default=None, description="Temporal Query"
    )
    lang: Optional[str] = Field(
        default=None, description="Language filter applied to the query"
    )
    model_config = ConfigDict(populate_by_name=True)



from typing import Union, Optional
from pydantic import BaseModel, Field, ConfigDict


class GeoQuery(BaseModel):
    """
    GeoQuery used for Subscriptions, as described in NGSI-LD Spec section 5.2.13
    """

    geometry: str = Field(
        description="A valid GeoJSON [8] geometry, type excepting GeometryCollection"
    )
    coordinates: Union[list, str] = Field(
        description="A JSON Array coherent with the geometry type as per "
        "IETF RFC 7946 [8]"
    )
    georel: str = Field(
        description="A valid geo-relationship as defined by clause 4.10 (near, "
        "within, etc.)"
    )
    geoproperty: Optional[str] = Field(
        default=None, description="Attribute Name as a short-hand string"
    )
    model_config = ConfigDict(populate_by_name=True)


def validate_ngsi_ld_query(q: str) -> str:
    """
    Valid query string as described in NGSI-LD Spec section 5.2.12
    Args:
        q: query string
    Returns:

    """
    return q



"""Vocabulary Models for Ontology Sources"""

import datetime
from typing import TYPE_CHECKING, List, Optional

from pydantic import ConfigDict, BaseModel, Field

from ...models.base import LogLevel

if TYPE_CHECKING:
    from . import Vocabulary, IdType, LoggingLevel


class DependencyStatement(BaseModel):
    """Information about one dependency statement in the source
    A dependency is a reference of one iri in an other entity definition
    """

    source_iri: str = Field(
        default="", description="Iri of the source containing the statement"
    )
    source_name: str = Field(
        default="", description="Name of the source containing the statement"
    )
    type: str = Field(
        description="Possible types: Parent Class, Relation Property, "
        "Relation Target"
    )
    class_iri: str = Field(description="Iri of the class containing the statement")
    dependency_iri: str = Field(description="Entity Iri of the dependency")
    fulfilled: bool = Field(
        description="True if the dependency_iri is registered in the " "vocabulary"
    )


class ParsingError(BaseModel):
    """Object represents one issue that arose while parsing a source,
    and holds all relevant details for that issue"""

    model_config = ConfigDict(use_enum_values=True)
    level: LogLevel = Field(description="Severity of error")
    source_iri: str = Field(description="Iri of the source containing the error")
    source_name: Optional[str] = Field(
        default=None, description="Name of the source, only set in get_function"
    )
    entity_type: str = Field(
        description="Type of the problematic entity: Class, Individual,.."
        "ID_type in string form"
    )
    entity_iri: str = Field(description="Iri of the problematic entity")
    entity_label: Optional[str] = Field(
        default=None, description="Name of the source, only set in get_function"
    )
    message: str = Field(description="Message describing the error")


class Source(BaseModel):
    """
    A source represent one file that was provided via file upload or link to the
    project and is parsed into the
    vocabulary
    """

    id: str = Field(default="", description="unique ID of the source; for internal use")
    source_name: str = Field(default="", description="Name of the source ")
    content: str = Field(
        default="", description="File content of the provided ontology file"
    )
    parsing_log: List["ParsingError"] = Field(
        default=[],
        description="Log containing all issues that were discovered while " "parsing",
    )
    dependency_statements: List[DependencyStatement] = Field(
        default=[], description="List of all statements in source"
    )
    timestamp: datetime.datetime = Field(
        description="timestamp when the source was added to the project"
    )
    ontology_iri: str = Field(
        default=None, description="Iri of the ontology of the source"
    )
    predefined: bool = Field(
        default=False,
        description="Stating if the source is a predefined source; "
        "a predefined source is added to each project containing "
        "owl:Thing and predefined Datatypes",
    )

    def get_number_of_id_type(self, vocabulary: "Vocabulary", id_type: "IdType") -> int:
        """Get the number how many entities of a given type are created by or
        influenced by this source

        Args:
            vocabulary (Vocabulary): Vocabulary of this project
            id_type (IdType): Idtype that should be counted

        Returns:
            int
        """

        from . import IdType

        id_func = "/"
        iri_list = []

        if id_type is IdType.class_:
            id_func = vocabulary.get_class_by_iri
            iri_list = vocabulary.classes
        elif id_type is IdType.object_property:
            id_func = vocabulary.get_object_property
            iri_list = vocabulary.object_properties
        elif id_type is IdType.data_property:
            id_func = vocabulary.get_data_property
            iri_list = vocabulary.data_properties
        elif id_type is IdType.individual:
            id_func = vocabulary.get_individual
            iri_list = vocabulary.individuals
        elif id_type is IdType.datatype:
            id_func = vocabulary.get_datatype
            iri_list = vocabulary.datatypes

        if id_func == "/":
            return -1

        counter = 0
        for iri in iri_list:
            entity = id_func(iri)
            if self.id in entity.source_ids:
                counter += 1
        return counter

    def get_name(self) -> str:
        """Get the name of the source

        Returns:
            str
        """
        return self.source_name

    def treat_dependency_statements(self, vocabulary: "Vocabulary"):
        """
        Log and purge all pointers/iris in entities that are not contained
        in the vocabulary

        Args:
            vocabulary (Vocabulary): Vocabulary of this project

        Returns:
            None
        """
        dependency_statements = []

        for class_ in vocabulary.get_classes():
            if self.id in class_.source_ids:
                dependency_statements.extend(
                    class_.treat_dependency_statements(vocabulary)
                )

        for individual_iri in vocabulary.individuals:
            individual = vocabulary.get_individual(individual_iri)
            if self.id in individual.source_ids:
                dependency_statements.extend(
                    individual.treat_dependency_statements(vocabulary)
                )

        for statement in dependency_statements:
            statement.source_iri = self.ontology_iri
            statement.source_name = self.source_name

        self.dependency_statements = dependency_statements

    def add_parsing_log_entry(
        self, level: "LoggingLevel", entity_type: "IdType", entity_iri: str, msg: str
    ):
        """
        Add a parsing log entry for an entity, if an issue in parsing
        was discovered

        Args:
            level (LoggingLevel): Logging level of the entry
            entity_type (IdType): Type of the enitity (Class, Individual,..)
            entity_iri (str): iri of the entity
            msg (str): message to display in log

        Returns:
            None
        """

        from . import ParsingError

        self.parsing_log.append(
            ParsingError(
                level=level,
                entity_type=str(entity_type),
                entity_iri=entity_iri,
                message=msg,
                source_iri=self.ontology_iri,
            )
        )

    def get_parsing_log(self, vocabulary: "Vocabulary") -> List["ParsingError"]:
        """Get the Parsinglog, where the labels of the entities are filled in

        Args:
            vocabulary (Vocabulary): Vocabulary of this project

        Returns:
            List[Dict[str, Union[LoggingLevel,'IdType',str]]]
        """
        for entry in self.parsing_log:

            entry.source_name = self.source_name
            try:
                label = vocabulary.get_label_for_entity_iri(entry.entity_iri)
                entry.entity_label = label
            except Exception:
                pass

        return self.parsing_log

    def clear(self):
        """Clear all logging and dependency data of the source

        Returns:
            None
        """
        self.parsing_log = []
        self.dependency_statements = []



"""Wrapper module to provide manipulation functions for vocabulary that
    should later be hidden from the user"""

import uuid
from enum import Enum


from pydantic import BaseModel, Field
from typing import Dict

from filip.models.base import LogLevel
from filip.semantics.vocabulary import *


class IdType(str, Enum):
    class_ = "Class"
    object_property = "Object Property"
    data_property = "Data Property"
    datatype = "Datatype"
    relation = "Relation"
    combined_relation = "Combined Relation"
    individual = "Individual"
    source = "Source"


class VocabularyBuilder(BaseModel):
    """Wrapper class to provide manipulation functions for vocabulary that
    should later be hidden from the user"""

    vocabulary: Vocabulary = Field(description="Vocabulary to manipulate")

    current_source: Source = Field(
        default=None,
        description="Current source to which entities are added,"
        "needed while parsing",
    )

    def clear(self):
        """Clear all objects form the vocabulary

        Returns:
            None
        """
        self.vocabulary.classes.clear()
        self.vocabulary.object_properties.clear()
        self.vocabulary.data_properties.clear()
        self.vocabulary.datatypes.clear()
        self.vocabulary.relations.clear()
        self.vocabulary.combined_object_relations.clear()
        self.vocabulary.combined_data_relations.clear()
        self.vocabulary.individuals.clear()
        self.vocabulary.id_types.clear()
        for source in self.vocabulary.sources.values():
            source.clear()

    def add_class(self, class_: Class):
        """Add a class to the vocabulary

        Args:
            class_ (Class): class to be added

        Returns:
            None
        """
        self._add_and_merge_entity(class_, self.vocabulary.classes, IdType.class_)

    def add_object_property(self, obj_prop: ObjectProperty):
        """Add an ObjectProperty to the vocabulary

        Args:
            obj_prop (ObjectProperty): ObjectProperty to be added

        Returns:
            None
        """
        self._add_and_merge_entity(
            obj_prop, self.vocabulary.object_properties, IdType.object_property
        )

    def add_data_property(self, data_prop: DataProperty):
        """Add an DataProperty to the vocabulary

        Args:
            data_prop (DataProperty): DataProperty to be added

        Returns:
            None
        """
        self._add_and_merge_entity(
            data_prop, self.vocabulary.data_properties, IdType.data_property
        )

    def add_datatype(self, datatype: Datatype):
        """Add a DataType to the vocabulary

        Args:
            datatype (Datatype): Datatype to be added

        Returns:
            None
        """
        self._add_and_merge_entity(datatype, self.vocabulary.datatypes, IdType.datatype)

    def add_predefined_datatype(self, datatype: Datatype):
        """Add a DataType to the vocabulary, that belongs to the source:
            Predefined

        Args:
            datatype (Datatype): Datatype to be added

        Returns:
            None
        """
        self.vocabulary.id_types[datatype.iri] = IdType.datatype
        self.vocabulary.datatypes[datatype.iri] = datatype
        datatype.predefined = True
        datatype.source_ids.add("PREDEFINED")

    def add_individual(self, individual: Individual):
        """Add an Individual to the vocabulary

        Args:
            individual (Individual): Individual to be added

        Returns:
            None
        """
        self._add_and_merge_entity(
            individual, self.vocabulary.individuals, IdType.individual
        )

    def add_relation_for_class(self, class_iri: str, rel: Relation):
        """Add a relation object to a class

        Args:
            class_iri: Iri of the class to which the relation should be added
            rel: Relation to add

        Returns:
            None
        """

        class_ = self.vocabulary.get_class_by_iri(class_iri)

        # for rel_id in class_.relation_ids:
        #     ex_rel = self.vocabulary.get_relation_by_id(rel_id)
        #     if rel.target_statement== rel.target_statement:
        #         return

        self.vocabulary.relations[rel.id] = rel
        class_.relation_ids.append(rel.id)
        self.vocabulary.id_types[rel.id] = IdType.relation

    def add_combined_object_relation_for_class(
        self, class_iri: str, crel: CombinedObjectRelation
    ):
        """Add a combined object relation object to a class

        Args:
            class_iri: Iri of the class to which the co-relation should be added
            crel: CombinedObjectRelation to add

        Returns:
            None
        """
        self.vocabulary.combined_object_relations[crel.id] = crel
        self.vocabulary.get_class_by_iri(class_iri).combined_object_relation_ids.append(
            crel.id
        )
        self.vocabulary.id_types[crel.id] = IdType.combined_relation

    def add_combined_data_relation_for_class(
        self, class_iri: str, cdata: CombinedDataRelation
    ):
        """Add a combined data relation object to a class

        Args:
            class_iri: Iri of the class to which the cd-relation should be added
            cdata: CombinedDataRelation to add

        Returns:
            None
        """
        self.vocabulary.combined_data_relations[cdata.id] = cdata
        self.vocabulary.get_class_by_iri(class_iri).combined_data_relation_ids.append(
            cdata.id
        )
        self.vocabulary.id_types[cdata.id] = IdType.combined_relation

    def add_source(self, source: Source, id: str = None):
        """Add a source to the vocabulary

        Args:
            source: source to add
            id: id of source, if none is given a random id is generated

        Returns:
            None
        """
        if id is None:
            source.id = uuid.uuid4().hex
        else:
            source.id = id
        self.vocabulary.id_types[source.id] = IdType.source
        self.vocabulary.sources[source.id] = source
        self.current_source = source

    def set_current_source(self, source_id: str):
        """Set the source of the vocabulary to which new added objects belong

        Args:
            source_id: id of source to activate

        Returns:
            None
        """
        assert source_id in self.vocabulary.sources
        self.current_source = self.vocabulary.sources[source_id]

    def _add_and_merge_entity(
        self, entity: Entity, entity_dict: Dict[str, Entity], id_type: IdType
    ):
        """Adds an entity to the vocabulary. If an entity with teh same iri
        already exists the label and comment are "merged" and both sources
        are noted

        Args:
            entity: Entity to check
            entity_dict: Existing entities
            id_type: Type of entity

        Raises:
            ParsingError: if Entity of iri exists but has a different IdType

        Returns:
            None
        """

        if entity.iri in self.vocabulary.id_types:
            if not id_type == self.vocabulary.id_types[entity.iri]:
                self.current_source.add_parsing_log_entry(
                    LogLevel.CRITICAL,
                    id_type,
                    entity.iri,
                    f"{entity.iri} from source "
                    f"{self.current_source.get_name()} "
                    f"exists multiple times in different catagories. It was "
                    f"only added for the category "
                    f"{self.vocabulary.id_types[entity.iri].value}",
                )
                return

            old_entity = entity_dict[entity.iri]

            def select_from(old: str, new: str, property: str) -> str:
                """
                Given two strings, one from the old_entity , one form the new
                one. It is selected which one to use.
                """
                if old == "":
                    return new
                elif new == "":
                    return ""
                else:
                    self.current_source.add_parsing_log_entry(
                        LogLevel.WARNING,
                        id_type,
                        entity.iri,
                        f"{property} from source "
                        f"{old_entity.get_source_names(self.vocabulary)} "
                        f"was overwritten",
                    )
                    return new

            entity.label = select_from(old_entity.label, entity.label, "label")
            entity.comment = select_from(old_entity.comment, entity.comment, "comment")

        self.vocabulary.id_types[entity.iri] = id_type
        entity.source_ids.add(self.current_source.id)
        entity_dict[entity.iri] = entity

    def entity_is_known(self, iri: str) -> bool:
        """Test if the given iri is in vocabulary, if not it belongs to a
        dependency which is not yet loaded

        Args:
            iri (str)

        Returns:
            bool
        """
        return iri in self.vocabulary.id_types



"""
The Simple Query Language provides a simplified syntax to retrieve entities
which match a set of conditions. A query is composed by a list of
statements separated by the ';' character. Each statement expresses a
matching condition. The query returns all the entities that match all
the matching conditions (AND logical operator).

For further details of the language please refer to:

https://telefonicaid.github.io/fiware-orion/api/v2/stable/
"""

import regex as re
from aenum import Enum
from typing import Union, List, Tuple, Any


class Operator(str, Enum):
    """
    The list of operators (and the format of the values they use) is as follows:
    """

    _init_ = "value __doc__"

    EQUAL = (
        "==",
        "Single element, e.g. temperature!=41. For an entity to "
        "match, it must contain the target property (temperature) "
        "and the target property value must not be the query value "
        "(41). "
        "A list of comma-separated values, e.g. color!=black,"
        "red. For an entity to match, it must contain the target "
        "property and the target property value must not be any "
        "of the values in the list (AND clause) (or not include any "
        "of the values in the list in case the target property "
        "value is an array). Eg. entities whose attribute color is "
        "set to black will not match, while entities whose "
        "attribute color is set to white will match."
        "A range, specified as a minimum and maximum separated by "
        ".., e.g. temperature!=10..20. For an entity to match, "
        "it must contain the target property (temperature) and the "
        "target property value must not be between the upper and "
        "lower limits (both included). Ranges can only be used "
        "with elements target properties that represent dates (in "
        "ISO8601 format), numbers or strings. ",
    )
    UNEQUAL = (
        "!=",
        "Single element, e.g. temperature!=41. For an entity to "
        "match, it must contain the target property "
        "(temperature) and the target property value must not be "
        "the query value (41). A list of comma-separated values, "
        "e.g. color!=black,red. For an entity to match, it must "
        "contain the target property and the target property "
        "value must not be any of the values in the list (AND "
        "clause) (or not include any of the values in the list "
        "in case the target property value is an array). Eg. "
        "entities whose attribute color is set to black will not "
        "match, while entities whose attribute color is set to "
        "white will match. A range, specified as a minimum and "
        "maximum separated by .., e.g. temperature!=10..20. For "
        "an entity to match, it must contain the target property "
        "(temperature) and the target property value must not be "
        "between the upper and lower limits (both included). "
        "Ranges can only be used with elements target properties "
        "that represent dates (in ISO8601 format), numbers or "
        "strings. ",
    )
    GREATER_THAN = (
        ">",
        "The right-hand side must be a single element, e.g. "
        "temperature>42. For an entity to match, it must "
        "contain the target property (temperature) and the "
        "target property value must be strictly greater than "
        "the query value (42). This operation is only valid "
        "for target properties of type date, number or "
        "string (used with target properties of other types "
        "may lead to unpredictable results). ",
    )
    LESS_THAN = (
        "<",
        "The right-hand side must be a single element, e.g. "
        "temperature<43. For an entity to match, it must "
        "contain the target property (temperature) and the "
        "target property value must be strictly less than the "
        "value (43). This operation is only valid for target "
        "properties of type date, number or string (used with "
        "target properties of other types may lead to "
        "unpredictable results). ",
    )
    GREATER_OR_EQUAL = (
        ">=",
        "The right-hand side must be a single element, "
        "e.g. temperature>=44. For an entity to match, "
        "it must contain the target property ("
        "temperature) and the target property value "
        "must be greater than or equal to that value "
        "(44). This operation is only valid for target "
        "properties of type date, number or string "
        "(used with target properties of other types "
        "may lead to unpredictable results). ",
    )
    LESS_OR_EQUAL = (
        "<=",
        "The right-hand side must be a single element, "
        "e.g. temperature<=45. For an entity to match, "
        "it must contain the target property (temperature) "
        "and the target property value must be less than "
        "or equal to that value (45). This operation is "
        "only valid for target properties of type date, "
        "number or string (used with target properties of "
        "other types may lead to unpredictable results). ",
    )
    MATCH_PATTERN = (
        "~=",
        "The value matches a given pattern, expressed as a "
        "regular expression, e.g. color~=ow. For an entity "
        "to match, it must contain the target property ("
        "color) and the target property value must match "
        "the string in the right-hand side, 'ow' in this "
        "example (brown and yellow would match, black and "
        "white would not). This operation is only valid "
        "for target properties of type string. ",
    )

    @classmethod
    def list(cls):
        """

        Returns:
            list of all valid values
        """
        return list(map(lambda c: c.value, cls))


class QueryStatement(Tuple):
    """
    Simple query statement
    """

    def __new__(cls, left: str, op: Union[str, Operator], right: Any):
        q = tuple.__new__(QueryStatement, (left, op, right))
        q = cls.validate(q)
        return q

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, value):
        """
        Validates statements

        Args:
            value:

        Returns:
        """
        if isinstance(value, (tuple, QueryStatement)):
            if len(value) != 3:
                raise TypeError("3-tuple required")
            if not isinstance(value[0], str):
                raise TypeError("First argument must be a string!")
            if value[1] not in Operator.list():
                raise TypeError("Invalid comparison operator!")
            if value[1] not in [
                Operator.EQUAL,
                Operator.UNEQUAL,
                Operator.MATCH_PATTERN,
            ]:
                try:
                    float(value[2])
                except ValueError as err:
                    err.args += (
                        "Invalid combination of operator and right " "hand side!",
                    )
                    raise
            return value
        elif isinstance(value, str):
            return cls.parse_str(value)
        raise TypeError

    def to_str(self):
        """
        Parses QueryStatement to String

        Returns:
            String
        """
        if not isinstance(self[2], str):
            right = str(self[2])
        elif self[2].isnumeric():
            right = f"{self[2]}"
        else:
            right = self[2]
        return "".join([self[0], self[1], right])

    @classmethod
    def parse_str(cls, string: str):
        """
        Generates QueryStatement form string

        Args:
            string:

        Returns:
            QueryStatement
        """
        for op in Operator.list():
            if re.fullmatch(rf"^\w((\w|[^&,?,/,#,\*,\s]\w)?)*{op}\w+$", string):
                args = string.split(op)
                if len(args) == 2:
                    if args[1].isnumeric():
                        try:
                            right = int(args[1])
                        except ValueError:
                            right = float(args[1])
                        return QueryStatement(args[0], op, right)
                    return QueryStatement(args[0], op, args[1])
                raise ValueError

    def __str__(self):
        """Return str(self)."""
        return self.to_str()

    def __repr__(self):
        """Return repr(self)."""
        return self.to_str().__repr__()


class QueryString:
    """
    Class for validated QueryStrings that can be used in api clients
    """

    def __init__(
        self, qs: Union[Tuple, QueryStatement, List[Union[QueryStatement, Tuple]]]
    ):
        qs = self.__check_arguments(qs=qs)
        self._qs = qs

    @classmethod
    def __check_arguments(cls, qs):
        """
        Check arguments on consistency

        Args:
            qs: queny statement object

        returns:
            List of QueryStatements
        """
        if isinstance(qs, List):
            for idx, item in enumerate(qs):
                if not isinstance(item, QueryStatement):
                    qs[idx] = QueryStatement(*item)
            # Remove duplicates
            qs = list(dict.fromkeys(qs))
        elif isinstance(qs, QueryStatement):
            qs = [qs]
        elif isinstance(qs, tuple):
            qs = [QueryStatement(*qs)]
        else:
            raise ValueError("Invalid argument!")
        return qs

    def update(self, qs: Union[Tuple, QueryStatement, List[QueryStatement]]):
        """
        Adds or updates QueryStatement within QueryString. First to arguments
        must match an existing argument for update. This redundant rules

        Args:
            qs: Query statement to add to the string object

        Returns:
            None
        """
        qs = self.__check_arguments(qs=qs)
        self._qs.extend(qs)
        self._qs = list(dict.fromkeys(qs))

    def remove(self, qs: Union[Tuple, QueryStatement, List[QueryStatement]]):
        """
        Remove Statement from QueryString
        Args:
            qs:

        Returns:

        """
        qs = self.__check_arguments(qs=qs)
        for q in qs:
            self._qs.remove(q)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        """validate QueryString"""
        if isinstance(v, QueryString):
            return v
        if isinstance(v, str):
            return cls.parse_str(v)
        raise ValueError("Invalid argument!")

    def to_str(self):
        """
        Parsing self.qs to string object

        Returns:
            String: query string that can be added to requests as parameter
        """
        return ";".join([q.to_str() for q in self._qs])

    @classmethod
    def parse_str(cls, string: str):
        """
        Creates QueryString from string

        Args:
            string:

        Returns:
            QueryString
        """
        q_parts = string.split(";")
        qs = []
        for part in q_parts:
            q = QueryStatement.parse_str(part)
            qs.append(q)
        return QueryString(qs=qs)

    def __str__(self):
        """Return str(self)."""
        return self.to_str()

    def __repr__(self):
        """Return repr(self)."""
        return self.to_str().__repr__()
